"""
Page controller example - for returning HTML pages
"""
from pyjolt.controller import Controller, get, path, produces
from pyjolt import MediaType, Request, Response

@path("/", open_api_spec=False) #not included in open api specs
class PageController(Controller):
    """
    Example controller for returning HTML pages
    """

    @get("/")
    @produces(MediaType.TEXT_HTML)
    async def index(self, req: Request) -> Response:
        """
        Example index page
        """
        #Context that is injected into the template
        context: dict = {
            "user": "John Doe"
        }
        return await req.res.html("index.html", context)

    @get("/about")
    @produces(MediaType.TEXT_HTML)
    async def about(self, req: Request) -> Response:
        """
        Example about page
        """
        context: dict = {
            "company": {
                "name": "My Company",
                "address": "1234 Main St, Anytown, USA"
            }
        }
        return await req.res.html("about.html", context)
        


"""
database module of pyjolt
"""
#re-export of some commonly used sqlalchemy objects 
#and methods for convenience.
from sqlalchemy import select, Select
from sqlalchemy.ext.asyncio import AsyncSession

from .sql_database import SqlDatabase
from .sqlalchemy_models import AsyncQuery, create_declerative_base

__all__ = ['SqlDatabase', 'select', 'Select',
           'AsyncSession', 'AsyncQuery', 'create_declerative_base']



"""
Utility methods for PyJolt
"""
import importlib
import importlib.util
import os
import re
import sys
import inspect
import asyncio
import mimetypes
import aiofiles
from asyncio import Task, Future
from typing import Callable, Any, Optional

from .exceptions import StaticAssetNotFound

def import_module(import_string: str):
    module_path, obj_name = import_string.split(":")
    try:
        module = importlib.import_module(module_path)
    except Exception as e:
        raise
    return getattr(module, obj_name)


def get_app_root_path(import_name: str) -> str:
    """
    Finds the root path of the application package on the file system or
    uses the current working directory
    """
    # First, check if the module is already imported and has a __file__ attribute.
    mod = sys.modules.get(import_name)
    if mod is not None and hasattr(mod, "__file__"):
        return os.path.dirname(os.path.abspath(mod.__file__)) # type: ignore

    # Tries to load the modules loader
    loader = importlib.util.find_spec(import_name)
    if loader is None or import_name == "__main__":
        return os.getcwd()

    # Checks if loader has a filename
    filepath = None
    if hasattr(loader, "get_filename"):
        filepath = loader.get_filename(import_name) # type: ignore

    # Tries to lookup the loaders path attribute
    if not filepath and hasattr(loader, "path"):
        filepath = loader.path # type: ignore

    if filepath is None:
        #Current working directory fallback
        return os.getcwd()

    # Return the directory name of the absolute path where the module resides.
    return os.path.dirname(os.path.abspath(filepath))

async def run_sync_or_async(func: Callable, *args, **kwargs):
    """
    Support for sync or async methods
    Runs async method directly or a sync method in a threadpool
    """
    if inspect.iscoroutinefunction(func):
        return await func(*args, **kwargs)

    loop = asyncio.get_running_loop()
    return await loop.run_in_executor(
        None,
        lambda: func(*args, **kwargs)
    )

def run_in_background(func: Callable[..., Any], *args, executor = None, **kwargs) -> Task|Future:
    """
    Fire-and-forget a function (async or sync) without awaiting its result.
    Useful for sending emails or other longer running tasks that do not
    need to wait for the result.
    """

    # current running event loop
    loop = asyncio.get_running_loop()

    if asyncio.iscoroutinefunction(func):
        # Schedule the async function to run
        return loop.create_task(func(*args, **kwargs))

    # If it's a sync function, run it in the default thread pool executor
    return loop.run_in_executor(executor, func, *args, **kwargs)

async def get_file(path: str, filename: Optional[str] = None, content_type: Optional[str] = None):
    """
    Asynchronously opens the file at `path`.
    - `filename` is optional (used for Content-Disposition).
    - `content_type` is optional (guess using `mimetypes` if not provided).
    
    Returns a tuple (status_code, headers, body_bytes).
    """

    # Guess the MIME type if none is provided
    guessed_type, _ = mimetypes.guess_type(path)
    content_type = content_type or (guessed_type or "application/octet-stream")

    headers = {
        "Content-Type": content_type
    }
    if filename:
        # For file download if filename is provided
        headers["Content-Disposition"] = f'attachment; filename="{filename}"'

    try:
        async with aiofiles.open(path, mode="rb") as f:
            data = await f.read()
    except FileNotFoundError:
        # pylint: disable-next=W0707,E0710
        raise StaticAssetNotFound()

    return 200, headers, data

async def get_range_file(res, file_path: str, range_header: str, content_type: str):
    """Returns a ranged response"""
    total = os.path.getsize(file_path)
    m = re.match(r"bytes=(\d+)-(\d*)", range_header)
    if not m:
        start, end, status = 0, total - 1, 200
    else:
        start = int(m.group(1))
        end   = int(m.group(2)) if m.group(2) else total - 1
        end   = min(end, total - 1)
        if start > end:
            raise StaticAssetNotFound()
        status = 206

    length = end - start + 1
    headers = {
        "Content-Type":   content_type,
        "Accept-Ranges":  "bytes",
        "Content-Length": str(length),
        "Cache-Control":  "public, max-age=300",
        "ETag": f'"{os.path.getmtime(file_path):.0f}-{length}"'
    }
    if status == 206:
        headers["Content-Range"] = f"bytes {start}-{end}/{total}"

    # **Don’t** read the bytes here.  Just stash info on `res`.  
    res.status(status)
    # merge headers onto res.headers
    res.headers.update(headers)
    # mark zero-copy parameters
    res.set_zero_copy({
        "file_path": file_path,
        "start":      start,
        "length":     length
    })
    return res



"""
Vector column for PostgreSQL database.
"""
from typing import Callable, List, Self, Type
from sqlalchemy.sql.elements import ColumnElement
from sqlalchemy.types import UserDefinedType

class Vector(UserDefinedType):
    """
    Custom Vector type for pgvector in PostgreSQL.

    Requires pgvector extension for the PostgreSQL database.
    """
    cache_ok = True

    def __init__(self, dimensions: int|None = None):
        self.dimensions: int|None = dimensions

    #pylint: disable-next=C0116
    def get_col_spec(self, **kw) -> str:
        if self.dimensions:
            return f"VECTOR({int(self.dimensions)})"
        return "VECTOR"

    #pylint: disable-next=C0116
    def bind_expression(self, bindvalue):
        return bindvalue
    
    #pylint: disable-next=C0116
    def coerce_compared_value(self, op, value):
        if isinstance(value, (list, tuple)):
            return self
        return super().coerce_compared_value(op, value)
    
    #pylint: disable-next=C0116
    def literal_processor(self, dialect) -> Callable:
        def process(value):
            if value is None:
                return "NULL"
            inside = ",".join(str(float(x)) for x in value)
            return f"'[{inside}]'::vector"
        return process
    
    #pylint: disable-next=C0116
    def compare_values(self, x: List, y: List) -> bool:
        if x is y:
            return True
        if x is None or y is None:
            return x is y
        if len(x) != len(y):
            return False
        eps = 1e-9
        return all(abs(float(a) - float(b)) <= eps for a, b in zip(x, y))
    
    class comparator_factory(UserDefinedType.Comparator):
        # distance operators exposed as methods for nicer query syntax
        def l2_distance(self, other):
            return self.expr.op("<->")(other)

        def inner_product(self, other):
            return self.expr.op("<#>")(other)

        def cosine_distance(self, other):
            return self.expr.op("<=>")(other)

    #pylint: disable-next=C0116
    def column_expression(self, colexpr) -> ColumnElement:
        return colexpr

    #pylint: disable-next=C0116
    def _with_collation(self, collation) -> Self:
        return self

    @property
    def python_type(self) -> Type[List]:
        return list



"""
Response class. Holds all information regarding responses to individual requests
"""
from typing import Any, Optional, TYPE_CHECKING, Self, TypeVar, Generic, Type

from .utilities import run_sync_or_async
from .http_statuses import HttpStatus

if TYPE_CHECKING:
    from .pyjolt import PyJolt

U = TypeVar("U")

class Response(Generic[U]):
    """
    Response class of application. Holds all data (headers, body, status_code of the response)
    Example return of route handler:
    ```
    return res.json({"message": "My message", "status": "some status"}).status(200)
    ```
    """
    def __init__(self, app: "PyJolt"):
        self._app = app
        self.status_code: int|HttpStatus = HttpStatus.OK #default status code is 200
        self.headers: dict = {}
        self.body: Optional[U] = None
        self.render_engine = self._app.jinja_environment
        self._zero_copy = None
        self._expected_body_type: Optional[Type[Any]] = None

    def status(self, status_code: int|HttpStatus) -> Self:
        """
        Sets status code of response
        """
        if isinstance(status_code, HttpStatus):
            self.status_code = status_code.value
        else:
            self.status_code = status_code
        return self
    
    def redirect(self, location: str,
                 status_code: int|HttpStatus = HttpStatus.SEE_OTHER) -> Self:
        """
        Redirects the client to the provided location
        """
        self.set_header("location", location)
        if isinstance(status_code, HttpStatus):
            self.status_code = status_code.value
        else:
            self.status_code = status_code
        
        self.body = None
        return self

    def json(self, data: Any) -> Self:
        """
        Sets data to response body and creates appropriate
        response headers. Sets default response status to 200
        """
        self.headers["content-type"] = "application/json"
        self.body = data
        return self

    def no_content(self) -> Self:
        """
        Returns a response with no content (body) and status 204
        """
        self.body = None
        self.status(204)
        return self

    def text(self, text: str) -> Self:
        """
        Creates text response with text/html content-type
        """
        self.body = text.encode("utf-8")
        self.status(HttpStatus.OK)
        return self
    
    async def html_from_string(self, text: str, context: Optional[dict[str, Any]] = None) -> Self:
        """
        Creates text response with text/html content-type
        """
        if context is None:
            context = {}
        for method in self.app.global_context_methods:
            additional_context = await run_sync_or_async(method)
            if not isinstance(additional_context, dict):
                raise ValueError("Return of global context method must be off type dictionary")
            context.update(additional_context)
        self.headers["content-type"] = "text/html"
        context["url_for"] = self.app.url_for
        rendered = await self.render_engine.from_string(text).render_async(**context)#self.render_engine.from_string(text).render(**context)
        #self.body = text.encode("utf-8")
        self.body = rendered.encode("utf-8")
        self.status(HttpStatus.OK)
        return self

    async def html(self, template_path: str, context: Optional[dict[str, Any]] = None) -> Self:
        """
        Renders html template and creates response with text/html content-type
        and default status code 200

        template_path: relative path of template inside the templates folder
        context: dictionary with data used in the template
        """

        if context is None:
            context = {}

        for method in self.app.global_context_methods:
            additional_context = await run_sync_or_async(method)
            if not isinstance(additional_context, dict):
                raise ValueError("Return of global context method must be off type dictionary")
            context = {**context, **additional_context}
        context["url_for"] = self.app.url_for

        template = self.render_engine.get_template(template_path)
        rendered = await template.render_async(**context)
        self.headers["content-type"] = "text/html"
        self.body = rendered.encode("utf-8")
        self.status(HttpStatus.OK)
        return self

    def send_file(self, body, headers) -> Self:
        """
        For sending files
        Sets correct headers and body of the response
        """
        for k, v in headers.items():
            self.headers[k] = v
        self.body = body
        return self

    def set_header(self, key: str, value: str) -> Self:
        """
        Sets or updates a header in the response.

        key: Header name
        value: Header value
        """
        self.headers[key.lower()] = value
        return self

    def set_cookie(self, cookie_name: str, value: str,
                   max_age: int|None = None, path: str = "/",
                   domain: str|None = None, secure: bool = False,
                   http_only: bool = True) -> Self:
        """
        Sets a cookie in the response.

        cookie_name: Cookie name
        value: Cookie value
        max_age: Max age of the cookie in seconds (optional)
        path: Path where the cookie is available (default "/")
        domain: Domain where the cookie is available (optional)
        secure: If True, the cookie is only sent over HTTPS (default False)
        http_only: If True, the cookie is inaccessible to JavaScript (default True) <-- MORE SECURE
        """
        cookie_parts = [f"{cookie_name}={value}"]

        if max_age is not None:
            cookie_parts.append(f"Max-Age={max_age}")
        if path:
            cookie_parts.append(f"Path={path}")
        if domain:
            cookie_parts.append(f"Domain={domain}")
        if secure:
            cookie_parts.append("Secure")
        if http_only:
            cookie_parts.append("HttpOnly")

        cookie_header = "; ".join(cookie_parts)
        if "set-cookie" in self.headers:
            self.headers["set-cookie"] += f", {cookie_header}"
        else:
            self.headers["set-cookie"] = cookie_header

        return self

    def delete_cookie(self, cookie_name: str,
                      path: str = "/", domain: Optional[str] = None) -> Self:
        """
        Deletes a cookie by setting its Max-Age to 0.

        cookie_name: Cookie name
        path: Path where the cookie was available (default "/")
        domain: Domain where the cookie was available (optional)
        """
        cookie_parts = [f"{cookie_name}=", "Max-Age=0", f"Path={path}"]

        if domain:
            cookie_parts.append(f"Domain={domain}")

        cookie_header = "; ".join(cookie_parts)
        if "set-cookie" in self.headers:
            self.headers["set-cookie"] += f", {cookie_header}"
        else:
            self.headers["set-cookie"] = cookie_header

        return self

    def set_zero_copy(self, data) -> Self:
        """Sets zero copy data for range responses"""
        self._zero_copy = data
        return self

    def _set_expected_body_type(self, t: Optional[Type[Any]]) -> None:
        self._expected_body_type = t

    def expected_body_type(self) -> Optional[Type[Any]]:
        return self._expected_body_type

    @property
    def zero_copy(self):
        """Returns zero copy data"""
        return self._zero_copy

    @property
    def app(self):
        """
        Returns application reference
        """
        return self._app



# migrations/env.py

from logging.config import fileConfig
from alembic import context
from sqlalchemy import engine_from_config, pool

# this is the Alembic Config object
config = context.config

# Interpret the config file for Python logging.
fileConfig(config.config_file_name)

# Obtain target_metadata from config
target_metadata = config.attributes.get("target_metadata", None)

def run_migrations_offline():
    """
    Run migrations in 'offline' mode.
    """
    url = config.get_main_option("sqlalchemy.url")
    context.configure(
        url=url,
        target_metadata=target_metadata,
        literal_binds=True,
        dialect_opts={"paramstyle": "named"},
    )
    with context.begin_transaction():
        context.run_migrations()

def run_migrations_online():
    """
    Run migrations in 'online' mode.
    """
    connectable = engine_from_config(
        config.get_section(config.config_ini_section),
        prefix="sqlalchemy.",
        poolclass=pool.NullPool,
    )

    with connectable.connect() as connection:
        context.configure(connection=connection, target_metadata=target_metadata)

        with context.begin_transaction():
            context.run_migrations()

if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



"""
In-memory cache implementation
"""
from typing import Optional, cast, TYPE_CHECKING
import asyncio

from cachetools import TTLCache

from .base_backend_cache import BaseCacheBackend

if TYPE_CHECKING:
    from ..pyjolt import PyJolt

class MemoryCacheBackend(BaseCacheBackend):
    """
    In-memory cache using cachetools.TTLCache for bounded size and base TTL.


    We honor per-item TTL by storing an explicit expire timestamp alongside
    the payload; TTLCache provides a global upper bound and eviction.
    """

    def __init__(self, default_ttl: int = 300, maxsize: int = 10_000):
        self.default_ttl = int(default_ttl)
        # Stores: key -> {payload: dict, expire: float}
        self._cache: TTLCache[str, dict] = TTLCache(maxsize=maxsize, ttl=self.default_ttl)

    # ---- config ----
    @classmethod
    def configure_from_app(cls, app: "PyJolt", variable_prefix: str) -> "MemoryCacheBackend":
        default_ttl = int(app.get_conf(f"{variable_prefix}CACHE_DURATION", 300))
        maxsize = int(app.get_conf(f"{variable_prefix}CACHE_MEMORY_MAXSIZE", 10_000))
        return cls(default_ttl=default_ttl, maxsize=maxsize)

    async def connect(self) -> None: # pragma: no cover - nothing to do
        return None

    async def disconnect(self) -> None:
        self._cache.clear()

    async def get(self, key: str) -> Optional[dict]:
        item = self._cache.get(key)
        if not item:
            return None
        expire = item.get("expire", 0)
        if expire < asyncio.get_event_loop().time():
            try:
                del self._cache[key]
            except KeyError:
                pass
            return None
        return cast(dict, item.get("payload"))

    async def set(self, key: str, value: dict, duration: Optional[int] = None) -> None:
        ttl = int(duration) if duration is not None else self.default_ttl
        expire = asyncio.get_event_loop().time() + ttl
        self._cache[key] = {"payload": value, "expire": expire}

    async def delete(self, key: str) -> None:
        try:
            del self._cache[key]
        except KeyError:
            pass

    async def clear(self) -> None:
        self._cache.clear()



"""
Exception controller implementation
"""
from functools import wraps
from typing import Callable, TYPE_CHECKING, Type

from ..controller.decorators import AsyncMethod, P, R
from ..controller.utilities import _extract_response_type
from ..utilities import run_sync_or_async

if TYPE_CHECKING:
    from ..pyjolt import PyJolt
    from ..response import Response
    from ..request import Request

class ExceptionHandler:

    def __init__(self, app: "PyJolt"):
        self._exception_mapping: dict[str, Callable] = {}
        self._app = app

    def get_exception_mapping(self) -> dict[str, Callable]:
        """Produces exception mapping"""
        owner_cls: "type[ExceptionHandler]|None" = self.__class__ or None
        handlers: dict[str, Callable] = {}
        if owner_cls is None:
            return handlers

        for name in dir(owner_cls):
            method = getattr(self, name)
            if not callable(method):
                continue
            handled_exceptions = getattr(method, "_handled_exceptions", []) or []
            for handled_exception in handled_exceptions:
                handlers[handled_exception.__name__] = method
            
        self._exception_mapping = handlers
        return handlers

    @property
    def app(self) -> "PyJolt":
        return self._app

def handles(*exceptions: Type[Exception]):
    """Decorator registers exceptions with handler method"""
    def decorator(func: Callable[P,R]) -> AsyncMethod:
        expected_body = _extract_response_type(func)
        @wraps(func)
        async def wrapper(self, *args: P.args, **kwargs: P.kwargs) -> "Response":
            # Request is auto-injected as first arg after self
            if not args:
                raise RuntimeError(
                    "Request must be auto-injected as the first argument after self."
                )
            req: "Request" = args[0] # type: ignore
            #pylint: disable-next=W0212
            req.response._set_expected_body_type(expected_body)
            res: "Response" = await run_sync_or_async(func, self, *args, **kwargs)
            return res
        handled_exceptions = getattr(func, "_handled_exceptions", []) or []
        handled_exceptions.extend(list(exceptions))
        #pylint: disable-next=W0212
        wrapper._handled_exceptions = handled_exceptions # type: ignore[attr-defined]
        return wrapper
    return decorator



"""
"""
from __future__ import annotations

from functools import wraps
from typing import Callable, Optional, Type, cast, TYPE_CHECKING

from ..utilities import run_sync_or_async
from ..base_extension import BaseExtension

from .base_backend_cache import BaseCacheBackend

if TYPE_CHECKING:
    from ..pyjolt import PyJolt
    from ..response import Response
    from ..request import Request

class Cache(BaseExtension):
    """
    Caching system for route handlers with **pluggable backend class**.

    Provide caching implementation as `CACHE_BACKEND` config. This should be
    a valid caching implementation (BaseCacheBackend)

    Default cache duration is set with `CACHE_DURATION` config (seconds)
    """

    def __init__(self, variable_prefix: str = ""):
        self._app: "Optional[PyJolt]" = None
        self._duration: int = 300
        self._backend: Optional[BaseCacheBackend] = None
        self._variable_prefix = variable_prefix

    def init_app(self, app: "PyJolt") -> None:
        self._app = app
        self._duration = int(self._app.get_conf("CACHE_DURATION", self._duration))
        pfx: str = self._variable_prefix
        backend_cls = self._app.get_conf(f"{pfx}CACHE_BACKEND", None)
        if backend_cls is None:
            #loads default backend - MemoryCacheBackend
            #pylint: disable-next=C0415
            from .memory_cache_backend import MemoryCacheBackend
            backend_cls = MemoryCacheBackend
        if not issubclass(backend_cls, BaseCacheBackend):
            raise TypeError("CACHE_BACKEND must be a class and subclass of BaseCacheBackend")

        self._backend = cast(Type[BaseCacheBackend], backend_cls).configure_from_app(app, pfx)

        self._app.add_extension(self)
        self._app.add_on_startup_method(self.connect)
        self._app.add_on_shutdown_method(self.disconnect)

    async def connect(self) -> None:
        if self._backend:
            await self._backend.connect()

    async def disconnect(self) -> None:
        if self._backend:
            await self._backend.disconnect()

    async def set(self, key: str, value: "Response", duration: Optional[int] = None) -> None:
        cached_value = {
            "status_code": value.status_code,
            "headers": value.headers,
            "body": value.body,
        }
        await cast(BaseCacheBackend, self._backend).set(key,
                                            cached_value, duration or self._duration)

    async def get(self, key: str, req: "Request") -> "Optional[Response]":
        payload = await cast(BaseCacheBackend, self._backend).get(key)
        if payload is None:
            return None
        return await self._make_cached_response(payload, req)

    async def delete(self, key: str) -> None:
        await cast(BaseCacheBackend, self._backend).delete(key)

    async def clear(self) -> None:
        await cast(BaseCacheBackend, self._backend).clear()

    async def _make_cached_response(self, cached_data: dict, req: "Request") -> "Response":
        req.res.body = cached_data["body"]
        req.res.status_code = cached_data["status_code"]
        req.res.headers = cached_data["headers"]
        return req.res

    def cache(self, duration: int = None) -> Callable:
        """Decorator for caching route handler results."""
        cache = self

        def decorator(handler: Callable) -> Callable:
            @wraps(handler)
            async def wrapper(self, *args, **kwargs) -> "Response":  # type: ignore[override]
                req: Request = args[0]
                method: str = req.method
                path: str = req.path
                query_params = sorted(req.query_params.items())
                cache_key = f"{handler.__name__}:{method}:{path}:{hash(frozenset(query_params))}"

                cached_value = await cache.get(cache_key, req)
                if cached_value is not None:
                    return cached_value

                res: Response = await run_sync_or_async(handler, self, *args, **kwargs)
                await cache.set(cache_key, res, duration)
                return res

            return wrapper

        return decorator



"""
App extensions
"""
from pyjolt.database import SqlDatabase
from pyjolt.database.migrate import Migrate
from pyjolt.caching import Cache

db: SqlDatabase = SqlDatabase()
migrate: Migrate = Migrate(db)
cache: Cache = Cache()

__all__ = ['db', 'migrate', 'cache']



"""
Request timing middleware
"""
from typing import TYPE_CHECKING
import time
from pyjolt.middleware import MiddlewareBase

if TYPE_CHECKING:
    from pyjolt.request import Request
    from pyjolt.response import Response

class TimingMW(MiddlewareBase):

    async def middleware(self, req: "Request") -> "Response":
        t0 = time.perf_counter()
        res = await self.next(req)# pass down
        #adds header with elapsed time (ms)
        res.headers["x-process-time-ms"] = str(int((time.perf_counter() - t0)*1000))
        return res



"""
CLI controller module for PyJolt.
"""
from typing import TYPE_CHECKING, Callable, cast, Type
from functools import wraps

from ..utilities import run_sync_or_async

if TYPE_CHECKING:
    from ..pyjolt import PyJolt

class CLIController:
    """
    Base class for CLI controllers.
    This class automatically registers methods decorated with @command and @argument
    as CLI commands in the provided PyJolt app instance.
    """
    def __init__(self, app: "PyJolt", cli_commands: dict):
        self._app: "PyJolt" = app
        self._cli_commands: dict = cli_commands
        self._register_commands()
    
    def _register_commands(self):
        for attr_name in dir(self):
            method = getattr(self, attr_name)
            if callable(method):
                command = getattr(method, "cli_command", {})
                if command.get("is_cli_command", False):
                    self._register_command(method, command)
    
    def _register_command(self, method: Callable, command: dict):

        cli_command = self._app.subparsers.add_parser(cast(str, command.get("command_name")),
                                                      help="Initialize the Alembic migration environment.")
        for arg_name, arg_info in command.get("arguments", {}).items():
            if arg_name.startswith("--"):
                cli_command.add_argument(arg_name, help=arg_info.get("description", ""))
            else:
                cli_command.add_argument(f"--{arg_name}", help=arg_info.get("description", ""))
        cli_command.set_defaults(func=lambda *args, **kwargs: self.run_command(method, command.get("arguments", {}), *args, **kwargs))
    
    def run_command(self, method: Callable, arg_info: dict[str, Type[int|float|str]], *args, **kwargs):
        for arg_name, info in arg_info.items():
            if arg_name in kwargs and "arg_type" in info:
                arg_type = info.get("arg_type", None)
                if not arg_type:
                    continue
                try:
                    kwargs[arg_name] = arg_type(kwargs[arg_name])
                except (ValueError, TypeError):
                    raise ValueError(f"Invalid type for argument '{arg_name}'. Expected {arg_type.__name__}.")
        return run_sync_or_async(method, *args, **kwargs)
    
    @property
    def app(self) -> "PyJolt":
        """Returns the PyJolt app instance."""
        return self._app

def command(command_name: str, help: str = ""):

    def decorator(func):
        @wraps(func)
        async def wrapper(self: "CLIController", *args, **kwargs):
            return await run_sync_or_async(func, self, *args, **kwargs)
        attr = getattr(wrapper, "cli_command", {})
        attr["is_cli_command"] = True
        attr["command_name"] = command_name
        attr["help"] = help
        arguments = attr.get("arguments", {})
        attr["arguments"] = arguments
        setattr(wrapper, "cli_command", attr)
        return wrapper
    return decorator

def argument(name: str, arg_type: Type[int|float|str], description: str = ""):
    """
    Argument decorator defines a command-line argument for a CLI command method.
    To use the argument in the command method, ensure the argument name matches the parameter name. Leading "--" are optional (are automatically added).
    The argument type can be int, float, or str. Input values are automatically converted to the specified type.
    Example:
    ```
        @command("greet", help="Greet a user with a message.")
        @argument("name", arg_type=str, description="The name of the user to greet.")
        async def greet(self, name: str):
            print(f"Hello, {name}!")
        
        @command("add", help="Adds two numbers")
        @argument("a", arg_type=int, description="First number")
        @argument("b", arg_type=int, description="Second number")
        async def add(self, a: int, b: int):
            print(f"{a} + {b} = {a+b}")
    ```
    Example usage:
        $ uv run app/cli.py greet --name Alice
        Hello, Alice!

        $ uv run app/cli.py add --a 2 --b 5
        2 + 5 = 7

    """
    def decorator(func):
        @wraps(func)
        async def wrapper(self: "CLIController", *args, **kwargs):
            return await run_sync_or_async(func, self, *args, **kwargs)
        attr = getattr(wrapper, "cli_command", {})
        arguments = attr.get("arguments", {})
        arguments[name] = {
            "description": description,
            "arg_type": arg_type
        }
        attr["arguments"] = arguments
        setattr(wrapper, "cli_command", attr)
        return wrapper
    return decorator



"""
All common http status codes
"""
from enum import IntEnum


class HttpStatus(IntEnum):
    # 1xx Informational
    CONTINUE = 100
    SWITCHING_PROTOCOLS = 101
    PROCESSING = 102              # WebDAV
    EARLY_HINTS = 103             # RFC 8297

    # 2xx Success
    OK = 200
    CREATED = 201
    ACCEPTED = 202
    NON_AUTHORITATIVE_INFORMATION = 203
    NO_CONTENT = 204
    RESET_CONTENT = 205
    PARTIAL_CONTENT = 206
    MULTI_STATUS = 207            # WebDAV
    ALREADY_REPORTED = 208        # WebDAV
    IM_USED = 226                 # RFC 3229

    # 3xx Redirection
    MULTIPLE_CHOICES = 300
    MOVED_PERMANENTLY = 301
    FOUND = 302
    SEE_OTHER = 303
    NOT_MODIFIED = 304
    USE_PROXY = 305
    TEMPORARY_REDIRECT = 307
    PERMANENT_REDIRECT = 308      # RFC 7538

    # 4xx Client Error
    BAD_REQUEST = 400
    UNAUTHORIZED = 401
    PAYMENT_REQUIRED = 402
    FORBIDDEN = 403
    NOT_FOUND = 404
    METHOD_NOT_ALLOWED = 405
    NOT_ACCEPTABLE = 406
    PROXY_AUTHENTICATION_REQUIRED = 407
    REQUEST_TIMEOUT = 408
    CONFLICT = 409
    GONE = 410
    LENGTH_REQUIRED = 411
    PRECONDITION_FAILED = 412
    PAYLOAD_TOO_LARGE = 413
    URI_TOO_LONG = 414
    UNSUPPORTED_MEDIA_TYPE = 415
    RANGE_NOT_SATISFIABLE = 416
    EXPECTATION_FAILED = 417
    IM_A_TEAPOT = 418             # RFC 2324
    MISDIRECTED_REQUEST = 421
    UNPROCESSABLE_ENTITY = 422    # WebDAV
    LOCKED = 423                  # WebDAV
    FAILED_DEPENDENCY = 424       # WebDAV
    TOO_EARLY = 425               # RFC 8470
    UPGRADE_REQUIRED = 426
    PRECONDITION_REQUIRED = 428  # RFC 6585
    TOO_MANY_REQUESTS = 429       # RFC 6585
    REQUEST_HEADER_FIELDS_TOO_LARGE = 431 # RFC 6585
    UNAVAILABLE_FOR_LEGAL_REASONS = 451   # RFC 7725

    # 5xx Server Error
    INTERNAL_SERVER_ERROR = 500
    NOT_IMPLEMENTED = 501
    BAD_GATEWAY = 502
    SERVICE_UNAVAILABLE = 503
    GATEWAY_TIMEOUT = 504
    HTTP_VERSION_NOT_SUPPORTED = 505
    VARIANT_ALSO_NEGOTIATES = 506 # RFC 2295
    INSUFFICIENT_STORAGE = 507    # WebDAV
    LOOP_DETECTED = 508           # WebDAV
    NOT_EXTENDED = 510            # RFC 2774
    NETWORK_AUTHENTICATION_REQUIRED = 511 # RFC 6585


"""
Exceptions submodule
"""

from .http_exceptions import (BaseHttpException,
                            StaticAssetNotFound,
                            AborterException,
                            MissingRequestData,
                            SchemaValidationError,
                            PydanticSchemaValidationError,
                            AuthenticationException,
                            UnauthorizedException,
                            InvalidJWTError,
                            abort,
                            html_abort)

from .runtime_exceptions import (CustomException,
                                 DuplicateRoutePath,
                                DuplicateExceptionHandler,
                                Jinja2NotInitilized,
                                MissingExtension,
                                MissingDependencyInjectionMethod,
                                MissingResponseObject,
                                MissingRouterInstance,
                                InvalidRouteHandler,
                                InvalidWebsocketHandler,
                                MethodNotControllerMethod,
                                UnexpectedDecorator)

from .exception_handler import ExceptionHandler, handles
from werkzeug.exceptions import NotFound, MethodNotAllowed

__all__ = ['CustomException',
            'BaseHttpException',
            'StaticAssetNotFound',
            'AborterException',
            'MissingRequestData',
            'SchemaValidationError',
            'PydanticSchemaValidationError',
            'AuthenticationException',
            'UnauthorizedException',
            "InvalidJWTError",
            'abort',
            'html_abort',
            'DuplicateRoutePath',
            'DuplicateExceptionHandler',
            'Jinja2NotInitilized',
            'MissingExtension',
            'MissingDependencyInjectionMethod',
            'MissingResponseObject',
            'MissingRouterInstance',
            'InvalidRouteHandler',
            'InvalidWebsocketHandler',
            'MethodNotControllerMethod',
            "UnexpectedDecorator",
            "ExceptionHandler",
            "handles",
            "NotFound",
            "MethodNotAllowed"]



"""
PyJolt application class
"""

# mypy: check-untyped-defs = True
import inspect
import argparse
import json
import asyncio
from typing import Any, Callable, Mapping, Optional, Type, TypeVar, cast
import aiofiles
from loguru import logger
from werkzeug.exceptions import NotFound, MethodNotAllowed
from pydantic import BaseModel

from jinja2 import (
    Environment,
    FileSystemLoader,
    select_autoescape,
    StrictUndefined,
    Undefined,
)

from .exceptions.http_exceptions import HtmlAborterException
from .http_statuses import HttpStatus
from .request import Request
from .response import Response
from .utilities import get_app_root_path, run_sync_or_async, import_module
from .exceptions import BaseHttpException, CustomException
from .router import Router
from .static import Static
from .open_api import OpenAPIController
from .controller import path

from .controller import Controller
from .exceptions import ExceptionHandler
from .base_extension import BaseExtension
from .configuration_base import BaseConfig
from .database.base_protocol import BaseModel as BaseModelClass
from .middleware import MiddlewareBase, AppCallableType
from .cli import CLIController

# ──────────────────────────────────────────────────────────────────────────────
# Monkey‐patch Uvicorn’s RequestResponseCycle.run_asgi so that, just before
# it invokes your ASGI app, it injects the real socket into the scope dict.
try:
    from uvicorn.protocols.http.h11_impl import RequestResponseCycle

    _orig_run_asgi = RequestResponseCycle.run_asgi

    async def _patched_run_asgi(self, application):
        # grab the raw socket from the transport and stash it into scope
        sock = None
        if hasattr(self, "transport") and self.transport is not None:
            sock = self.transport.get_extra_info("socket")
        if sock is not None:
            self.scope["socket"] = sock

        # now call the real ASGI loop
        return await _orig_run_asgi(self, application)

    RequestResponseCycle.run_asgi = _patched_run_asgi
# pylint: disable-next=W0718
except Exception as e:
    logger.debug(
        "Could not patch RequestResponseCycle.run_asgi; "
        "os.sendfile() zero-copy will fall back to aiofiles. "
        f"Patch error: {e}"
    )
# ──────────────────────────────────────────────────────────────────────────────


PYJOLT_ASCIART: str = r"""
  _______     __  _  ____  _   _______ 
 |  __ \ \   / / | |/ __ \| | |__   __|
 | |__) \ \_/ /  | | |  | | |    | |   
 |  ___/ \   /   | | |  | | |    | |   
 | |      | | |__| | |__| | |____| |   
 |_|      |_|\____/ \____/|______|_|   
A Fast, Simple, and Productive Python Web Framework
"""

PYJOLT_VERSION: str = "0.9.x"

T = TypeVar("T", bound="PyJolt")

def app_path(url_path: Optional[str] = None) -> Callable[[Type[T]], Type[T]]:
    def decorator(cls: Type[T]) -> Type[T]:
        setattr(cls, "_base_url_path", url_path)
        return cls

    return decorator


def app(import_name: str, configs: Type[BaseConfig]) -> Callable[[Type[T]], Type[T]]:
    def decorator(cls: Type[T]) -> Type[T]:
        setattr(cls, "_app_configs", {"import_name": import_name, "configs": configs})
        return cls

    return decorator


def on_startup(func) -> Callable:
    """
    Decorated methods will run in alphabetical order on app startup
    """
    setattr(func, "_on_startup_method", True)
    return func


def on_shutdown(func) -> Callable:
    """
    Decorated methods will run in alphabetical order on app shutdown
    """
    setattr(func, "_on_shutdown_method", True)
    return func


class MissingAppConfigurations(Exception):
    def __init__(
        self,
        msg: str = ("Missing application configurations. "
                    "Please make sure to use the @app_configs "
                    "decorator with appropriate arguments.")
    ):
        super().__init__(msg)


class MissingImportModule(Exception):
    def __init__(self, msg: str):
        super().__init__(msg)


class WrongModuleLoadType(Exception):
    def __init__(self, msg: str):
        super().__init__(msg)


def validate_config(config_obj_or_type: Type[BaseConfig]) -> BaseConfig:
    # If it's already an instance
    if isinstance(config_obj_or_type, BaseConfig):
        return config_obj_or_type  # already validated by Pydantic

    if inspect.isclass(config_obj_or_type) and issubclass(
        config_obj_or_type, BaseConfig
    ):
        try:
            instance = config_obj_or_type()
        except Exception as e:
            raise MissingAppConfigurations(
                f"Could not instantiate config class {config_obj_or_type.__name__}: {e}"
            ) from e
        return BaseConfig.model_validate(instance.model_dump())

    raise MissingAppConfigurations("Configs must be a subclass of pyjolt.BaseConfig.")


class PyJolt:
    """PyJolt class implementation. Used to create a new application instance"""

    def __init__(self):
        """Init function"""
        app_configs: dict[str, str | object | dict] | None = getattr(
            self.__class__, "_app_configs", None
        )
        if app_configs is None:
            raise MissingAppConfigurations()

        import_name: str = cast(str, app_configs.get("import_name", None))
        configs: Type[BaseConfig] = cast(
            Type[BaseConfig], app_configs.get("configs", None)
        )
        if configs is None or not issubclass(configs, BaseConfig):
            raise MissingAppConfigurations(
                "Missing valid configs object in @app_configs. Configuration class must inherit from pyjolt.BaseConfig"
            )

        self._app_base_url: str = getattr(self.__class__, "_base_url_path", "")
        self._is_built = False
        self._root_path = get_app_root_path(import_name)
        # Dictionary which holds application configurations
        validated_configs: BaseConfig = validate_config(configs)
        self._configs = {**validated_configs.model_dump()}
        self._static_files_path = f"{self._root_path + self.get_conf('STATIC_DIR')}"
        self._templates_path = self._root_path + self.get_conf("TEMPLATES_DIR")

        self._url_for_alias: dict[str, str] = {
            self.get_conf("STATIC_CONTROLLER_NAME"): "Static.get"
        }

        #creates Jinja2 environment for entire app
        self._jinja_environment = Environment(
            loader=FileSystemLoader(self._templates_path),
            autoescape=select_autoescape(["html", "xml"]),
            undefined=StrictUndefined
            if self.get_conf("TEMPLATES_STRICT", True)
            else Undefined,
            auto_reload=self.get_conf("DEBUG", False),
            enable_async=True,
        )

        self._router = Router(self.get_conf("STRICT_SLASHES", False))
        self._logger = logger

        self._app: AppCallableType = self._base_app
        self._middleware: list[Callable] = []
        self._controllers: dict[str, "Controller"] = {}
        self._cli_controllers: dict[str, "CLIController"] = {}
        self._exception_handlers: dict[str, Callable] = {}
        self._json_spec: Optional[dict] = None
        self._db_models: dict[str, list[BaseModelClass]] = {}

        self._extensions: dict = {}
        self.global_context_methods: list[Callable] = []

        self._on_startup_methods: list[Callable] = []
        self._on_shutdown_methods: list[Callable] = []

        self._get_startup_methods()
        self._get_shutdown_methods()

        self.cli = argparse.ArgumentParser(description="PyJolt CLI")
        self.subparsers = self.cli.add_subparsers(dest="command", help="CLI commands")
        self.cli_commands: dict = {}

        models: Optional[list[str]] = self.get_conf("MODELS", None)
        controllers: Optional[list[str]] = self.get_conf("CONTROLLERS", None)
        cli_controllers: Optional[list[str]] = self.get_conf("CLI_CONTROLLERS", None)
        exception_handlers: Optional[list[str]] = self.get_conf(
            "EXCEPTION_HANDLERS", None
        )
        extensions: Optional[list[str]] = self.get_conf("EXTENSIONS", None)
        middleware: Optional[list[str]] = self.get_conf("MIDDLEWARE", None)
        self._load_modules(models)
        self._load_modules(extensions)
        self._load_modules(controllers)
        self._load_modules(cli_controllers)
        self._load_modules(exception_handlers)
        self._load_modules(middleware)

    def _load_modules(self, modules: Optional[list[str]] = None):
        if modules is None:
            return
        for import_string in modules:
            obj = import_module(import_string)
            if obj is None:
                raise MissingImportModule(
                    f"Failed to load module: {import_string}. Check path in configurations."
                )
            if inspect.isclass(obj) and issubclass(obj, Controller):
                self.logger.info(f"Registering controller: {obj.__name__}")
                self.register_controller(obj)
                continue
            if inspect.isclass(obj) and issubclass(obj, ExceptionHandler):
                self.logger.info(f"Registering exception handler: {obj.__name__}")
                self.register_exception_handler(obj)
                continue
            if isinstance(obj, BaseExtension):
                self.logger.info(f"Initilizing extension: {obj.__class__.__name__}")
                obj.init_app(self)
                continue
            if inspect.isclass(obj) and issubclass(obj, BaseModelClass):
                self.logger.info(f"Loaded database model: {obj.__name__}")
                if obj.db_name() not in self._db_models:
                    self._db_models[obj.db_name()] = []
                self._db_models[obj.db_name()].append(cast(BaseModelClass, obj))
                continue
            if inspect.isclass(obj) and issubclass(obj, CLIController):
                self.logger.info(f"Registering cli controller: {obj.__name__}")
                commands = getattr(obj, "_cli_command", {})
                cli_controller = obj(self, commands)
                self._cli_controllers[obj.__name__] = cli_controller
                continue
            if inspect.isclass(obj) and issubclass(obj, MiddlewareBase):
                self.logger.info(f"Registering middleware: {obj.__name__}")
                self._middleware.append(
                    lambda app, next_app, mdlwr_class=obj: mdlwr_class(app, next_app)
                )
                continue
            raise WrongModuleLoadType(
                f"Failed to load module {obj.__name__ or obj.__class__.__name__}. Extensions must be passed as instances, controllers, cli controllers, exception handlers and middleware as classes."
            )

    def _get_startup_methods(self):
        methods = []
        for name in dir(self):
            method = getattr(self, name)
            if callable(method) and getattr(method, "_on_startup_method", False):
                methods.append((name, method))
        methods.sort(key=lambda x: x[0])  # by method name
        self._on_startup_methods = [m for _, m in methods]

    def _get_shutdown_methods(self):
        methods = []
        for name in dir(self):
            method = getattr(self, name)
            if callable(method) and getattr(method, "_on_shutdown_method", False):
                methods.append((name, method))
        methods.sort(key=lambda x: x[0])  # by method name
        self._on_shutdown_methods = [m for _, m in methods]

    def get_conf(self, config_name: str, default: Any = None) -> Any:
        """
        Returns app configuration with provided config_name.
        Raises error if configuration is not found.
        """
        return self.configs.get(config_name, default)

    def add_global_context_method(self, func: Callable):
        """
        Adds global context method to global_context_methods array
        """
        self.global_context_methods.append(func)

    async def _base_app(self, req: Request) -> Response:
        """
        The bare-bones application without any middleware.
        Calls the route handler directly.
        """
        res: Response = await run_sync_or_async(
            req.route_handler, req, **req.route_parameters
        )
        return res

    async def abort_route_not_found(self, send, req: Request, path_data: Mapping[str, Any]):
        """
        Aborts request because route was not found
        """
        exc: NotFound|MethodNotAllowed|None = path_data.get("exc")
        if exc is not None:
            handler: Callable|None = (
                self._exception_handlers.get(exc.__class__.__name__, None) or None
            )
            if handler:
                res = await run_sync_or_async(handler, req, exc)
                response_type = res.expected_body_type() or exc.__class__
                return await self.send_response(res, send, response_type)
        ##sends generic response if custom handler not available
        await send(
            {
                "type": "http.response.start",
                "status": 404,
                "headers": [(b"content-type", b"application/json")],
            }
        )
        await send(
            {
                "type": "http.response.body",
                "body": b'{ "status": "error", "message": "Endpoint not found" }',
            }
        )

    async def send_response(
        self, res: Response, send, response_type: Optional[Type[Any]] = None
    ):
        """
        Sends response
        """
        # Build headers for ASGI send
        headers = []
        for k, v in res.headers.items():
            headers.append((k.encode("utf-8"), v.encode("utf-8")))
        await send(
            {
                "type": "http.response.start",
                "status": res.status_code.value
                if isinstance(res.status_code, HttpStatus)
                else res.status_code,
                "headers": headers,
            }
        )
        # Zero-copy _parameters_ were stashed in res._zero_copy
        if res.zero_copy is not None:
            params = res.zero_copy
            file_path = params["file_path"]
            start = params["start"]
            length = params["length"]

            # stream in 1 MiB chunks
            chunk_size = 1 * 1024 * 1024
            remaining = length

            async with aiofiles.open(file_path, "rb") as f:
                await f.seek(start)
                while remaining > 0:
                    to_read = min(remaining, chunk_size)
                    chunk = await f.read(to_read)
                    if not chunk:
                        break
                    remaining -= len(chunk)
                    await send(
                        {
                            "type": "http.response.body",
                            "body": chunk,
                            "more_body": remaining > 0,
                        }
                    )
            return
        if (
            response_type
            and issubclass(
                response_type, (CustomException, BaseHttpException, Exception)
            )
            and not issubclass(response_type, HtmlAborterException)
        ):
            res.body = json.dumps(res.body).encode("utf-8")
        elif (
            response_type
            and res.body
            and issubclass(response_type, BaseModel)
            and isinstance(res.body, dict)
        ):
            res.body = response_type(**res.body).model_dump_json().encode("utf-8")
        elif (
            response_type
            and res.body
            and issubclass(response_type, BaseModel)
            and isinstance(res.body, BaseModel)
        ):
            res.body = res.body.model_dump_json().encode("utf-8")
        elif res.body and response_type is None and isinstance(res.body, BaseModel):
            res.body = res.body.model_dump_json().encode("utf-8")
        elif res.body and not isinstance(res.body, (bytes, bytearray)):
            res.body = json.dumps(res.body).encode("utf-8")

        await send(
            {
                "type": "http.response.body",
                "body": res.body or b"",
            }
        )

    async def _lifespan_app(self, _, receive, send):
        """This loop will listen for 'startup' and 'shutdown'"""
        while True:
            message = await receive()

            if message["type"] == "lifespan.startup":
                # Run all your before_start methods once
                for method in self._on_startup_methods:
                    await run_sync_or_async(method)

                # Signal uvicorn that startup is complete
                await send({"type": "lifespan.startup.complete"})

            elif message["type"] == "lifespan.shutdown":
                # Run your after_start methods (often used for cleanup)
                for method in self._on_shutdown_methods:
                    await run_sync_or_async(method)

                # Signal uvicorn that shutdown is complete
                await send({"type": "lifespan.shutdown.complete"})
                return  # Exit the lifespan loop

    async def _handle_http_request(self, scope, receive, send):
        """
        Handles http requests with CORS support (refactored)
        """
        method: str = scope["method"].upper()
        url_path: str = scope["path"]
        self._log_request(scope, method, url_path)

        route_handler, path_kwargs = self.router.match(url_path, method)
        req = Request(scope, receive, self, path_kwargs, cast(Callable, route_handler))

        if not route_handler:
            return await self.abort_route_not_found(send, req, path_kwargs)

        # CORS handling
        cors_opts = self._resolve_cors_options(cast(Callable, route_handler))
        origin = self._get_header(scope, "origin")
        is_preflight = (method == "OPTIONS") and bool(self._get_header(scope, "access-control-request-method"))

        if cors_opts["enabled"] and origin:
            if not self._origin_allowed(origin, cors_opts["allow_origins"]):
                return await self._cors_forbidden(send)

            if is_preflight:
                return await self._handle_preflight(scope, send, origin, cors_opts)

            # Wrap send to inject CORS headers on normal responses
            send = self._wrap_cors_send(send, origin, cors_opts)

        try:
            try:
                res: Response = await self._app(req)
                response_type: Optional[Type[Any]] = req.response.expected_body_type()
                return await self.send_response(res, send, response_type)
            except HtmlAborterException as exc:
                res = (await req.res.html(exc.template, context=exc.data)).status(
                    exc.status_code
                )
                return await self.send_response(res, send, None)
            # pylint: disable-next=W0718
            except Exception as exc:
                handler = (
                    self._exception_handlers.get(exc.__class__.__name__, None) or None
                )
                if not handler:
                    raise Exception("Unhandled exception occured") from exc
                res = await run_sync_or_async(handler, req, exc)
                response_type = res.expected_body_type() or exc.__class__
                return await self.send_response(res, send, response_type)
        # pylint: disable-next=W0718
        except Exception as exc:
            # Catches every error and returns internal server error message
            # if the app is in production (DEBUG = False)
            # else reraises the error
            if not self.get_conf("DEBUG", False):
                res = req.res.json(
                    {
                        "status": "error",
                        "message": "Internal server error",
                    }
                ).status(HttpStatus.INTERNAL_SERVER_ERROR)
                self.logger.critical(f"Unhandled critical error: ({req.method}) {req.path}, {req.route_parameters}")
                return await self.send_response(res, send, Exception)
            raise


    def _log_request(self, scope, method: str, url_path: str) -> None:
        """
        Logs incoming request
        """
        logger.info(
            f"HTTP request. CLIENT: {(scope.get('client') or ("-", ""))[0]}, SCHEME: {scope['scheme']}, METHOD: {method}, PATH: {url_path}, QUERY_STRING: {scope['query_string'].decode('utf-8')}"
        )

    def register_static_controller(self, base_path: str):
        static_controller_dec = path(f"{base_path}", open_api_spec=False)
        static_controller = static_controller_dec(Static)
        self.register_controller(static_controller)  # type: ignore

    def register_openapi_controller(self):
        openapi_controller_dec = path(
            self.get_conf("OPEN_API_URL"), open_api_spec=False
        )
        openapi_controller = openapi_controller_dec(OpenAPIController)
        self.register_controller(openapi_controller)

    def build(self) -> None:
        """
        Build the final app by wrapping self._app in all middleware.
        Apply them in reverse order so the first middleware in the list
        is the outermost layer.
        """
        print(PYJOLT_ASCIART)
        print(f"Starting PyJolt {PYJOLT_VERSION} application '{self.app_name}'")
        self.register_static_controller(self.get_conf("STATIC_URL"))
        if self.get_conf("OPEN_API", False):
            self.build_openapi_spec()
            self.register_openapi_controller()
        app: AppCallableType = self._base_app
        for factory in reversed(self._middleware):
            app = factory(self, app)
        self._app = app
        self._is_built = True

    def add_extension(self, extension):
        """
        Adds extension to extension map
        """
        ext_name: str = (
            extension.__name__
            if hasattr(extension, "__name__")
            else extension.__class__.__name__
        )
        self._extensions[ext_name] = extension

    def activate_extension(self, extension: "Type[BaseExtension]"):
        extension_instance = extension()
        extension_instance.init_app(self)

    def _add_route_function(
        self, method: str, url_path: str, func: Callable, endpoint_name: str
    ):
        """
        Adds the function to the Router.
        Raises DuplicateRoutePath if a route with the same (method, path) is already registered.
        """
        try:
            self.router.add_route(url_path, func, [method], endpoint_name)
        except Exception as e:
            # Detect more specific errors?
            raise e

    def register_controller(self, *ctrls: "type[Controller]"):
        """Registers controller class with application"""
        for ctrl in ctrls:
            ctrl_path: str = getattr(ctrl, "_controller_path")
            ctrl_open_api_spec = getattr(ctrl, "_include_open_api_spec")
            ctrl_open_api_tags = getattr(ctrl, "_open_api_tags", None)
            ctrl_instance = ctrl(
                self, ctrl_path, ctrl_open_api_spec, ctrl_open_api_tags
            )

            self._controllers[ctrl_instance.path] = ctrl_instance
            endpoint_methods: dict[str, dict[str, str | Callable]] = (
                ctrl_instance.get_endpoint_methods()
            )
            for http_method, endpoints in endpoint_methods.items():
                for url_path, method in endpoints.items():
                    method_name: Callable = method["method"].__name__  # type: ignore
                    # handler: Callable = getattr(ctrl_instance, method_name) # type: ignore
                    endpoint_name: str = (
                        f"{ctrl_instance.__class__.__name__}.{method_name}"
                    )
                    self._add_route_function(
                        http_method,
                        self._app_base_url + ctrl_instance.path + url_path,
                        method["method"],
                        endpoint_name,
                    )

    def register_exception_handler(self, *handlers: "type[ExceptionHandler]"):
        """Registers exception controller with application"""
        for handler in handlers:
            handler_instance = handler(self)
            handled_exceptions = handler_instance.get_exception_mapping()
            self._exception_handlers.update(handled_exceptions)

    def url_for(self, endpoint: str, **values) -> str:
        """
        Returns url for endpoint method
        :param endpoint: the name of the endpoint handler method namespaced
        with the blueprint name (if in blueprint)
        :param values: dynamic route parameters
        :return: url (string) for endpoint
        """
        endpoint = self._url_for_alias.get(endpoint, endpoint)
        adapter = self.router.url_map.bind("")  # Binds map to base url
        # If a value starts with a forward slash, systems like MacOS/Linux treat it as an absolute path
        try:
            return adapter.build(endpoint, values)
        except NotFound as exc:
            raise ValueError(f"Endpoint '{endpoint}' does not exist.") from exc
        except MethodNotAllowed as exc:
            raise ValueError(
                f"Endpoint '{endpoint}' exists but does not allow the method."
            ) from exc
        except Exception as exc:
            raise ValueError(
                f"Error building URL for endpoint '{endpoint}': {exc}"
            ) from exc

    def build_openapi_spec(self):
        """Builds open api spec"""
        from .open_api import build_openapi

        self._json_spec = build_openapi(
            self._controllers,
            title=self.app_name,
            version=self.version,
            openapi_version="3.0.3",
            servers=["http://localhost:8080"],
        )

    def add_on_startup_method(self, func: Callable):
        """
        Adds method to on_startup collection
        """
        self._on_startup_methods.append(func)

    def add_on_shutdown_method(self, func: Callable):
        """
        Adds method to on_shutdown collection
        """
        self._on_shutdown_methods.append(func)
    
    def register_alias(self, alias: str, endpoint: str):
        """
        Registers an alias for an endpoint name.
        Useful for url_for lookups.
        """
        self._url_for_alias[alias] = endpoint

    def run_cli(self):
        """
        Executes the registered CLI commands.
        """
        args = self.cli.parse_args()
        if hasattr(args, "func"):
            func_args = args._get_args()
            func_kwargs = {name: value for name, value in args._get_kwargs() if name not in ["command", "func"]}
            asyncio.run(args.func(*func_args, **func_kwargs))  # pass the parsed arguments object
        else:
            self.cli.print_help()

    @property
    def json_spec(self) -> dict | None:
        return self._json_spec

    @property
    def router(self) -> Router:
        """Router instance property of the app"""
        return self._router

    @property
    def configs(self) -> dict[str, Any]:
        """
        Returns configuration dictionary
        """
        return self._configs

    @property
    def global_context(self):
        """
        Decorator registers method as a context provider for html templates.
        The return of the decorated function should be dictionary with key-value pairs.
        The returned dictionary is added to the context of the render_template method
        """

        def decorator(func: Callable):
            self.add_global_context_method(func)
            return func

        return decorator

    @property
    def root_path(self) -> str:
        """
        Returns root path of application
        """
        return self._root_path

    @property
    def app(self):
        """
        Returns self
        For compatibility with the Controller class
        which contains the app object on the app property
        """
        return self

    @property
    def static_files_path(self) -> str:
        """Static files paths"""
        return self._static_files_path

    @property
    def version(self) -> str:
        return self.get_conf("VERSION")

    @property
    def app_name(self) -> str:
        return self.get_conf("APP_NAME")

    @property
    def logger(self):
        return self._logger
    
    @property
    def jinja_environment(self) -> Environment:
        return self._jinja_environment

    async def __call__(self, scope, receive, send):
        """
        Once built, __call__ just delegates to the fully wrapped app.
        """
        if not self._is_built:
            self.build()
        if scope["type"] == "lifespan":
            return await self._lifespan_app(scope, receive, send)
        if scope["type"] == "http":
            return await self._handle_http_request(scope, receive, send)
        raise ValueError(f"Unsupported scope type {scope['type']}")
    
    # ───────────────────────── CORS utilities ─────────────────────────

    def _get_header(self, scope, name: str) -> Optional[str]:
        target = name.lower().encode("latin1")
        for k, v in scope.get("headers", []):
            if k.lower() == target:
                try:
                    return v.decode("latin1")
                except Exception:
                    return None
        return None

    def _normalize_list(self, v, *, upper: bool = False) -> list[str]:
        if v is None:
            return []
        if isinstance(v, str):
            v = [v]
        out = [s.strip() for s in v if s and s.strip()]
        if upper:
            out = [s.upper() for s in out]
        return out

    def _global_cors_options(self) -> dict[str, Any]:
        return {
            "enabled": self.get_conf("CORS_ENABLED"),
            "allow_origins": self.get_conf("CORS_ALLOW_ORIGINS"),
            "allow_methods": self.get_conf("CORS_ALLOW_METHODS"),
            "allow_headers": self.get_conf("CORS_ALLOW_HEADERS"),
            "expose_headers": self.get_conf("CORS_EXPOSE_HEADERS"),
            "allow_credentials": self.get_conf("CORS_ALLOW_CREDENTIALS"),
            "max_age": self.get_conf("CORS_MAX_AGE"),
        }

    def _resolve_cors_options(self, handler: Optional[Callable]) -> dict[str, Any]:
        """
        Merge global and endpoint-level CORS options.
        If handler has @no_cors, CORS is completely disabled.
        """
        opts = self._global_cors_options()

        if handler is None:
            return opts

        # Check if handler explicitly disables CORS
        if getattr(handler, "_disable_cors", False):
            opts["enabled"] = False
            return opts

        # Merge @cors decorator overrides
        overrides = getattr(handler, "_cors_options", None) or {}
        for k, v in overrides.items():
            if v is not None:
                opts[k] = v

        opts["allow_origins"] = self._normalize_list(opts["allow_origins"])
        opts["allow_methods"] = self._normalize_list(opts["allow_methods"], upper=True)
        opts["allow_headers"] = self._normalize_list(opts["allow_headers"])
        opts["expose_headers"] = self._normalize_list(opts["expose_headers"])
        return opts

    def _origin_allowed(self, origin: Optional[str], allow_origins: list[str]) -> bool:
        if not origin:
            return True
        if "*" in allow_origins:
            return True
        return origin in allow_origins

    def _build_cors_headers_for_preflight(
        self,
        *,
        origin: str,
        request_method: Optional[str],
        request_headers: Optional[str],
        opts: dict[str, Any]
    ) -> list[tuple[bytes, bytes]]:
        allow_methods = opts["allow_methods"]
        allow_headers = opts["allow_headers"]

        # echo requested headers if none configured
        if not allow_headers and request_headers:
            allow_headers = [h.strip() for h in request_headers.split(",") if h.strip()]

        # If "*" configured but credentials allowed: echo back the Origin
        allow_origin_value = "*"
        if opts["allow_credentials"] or ("*" not in opts["allow_origins"]):
            allow_origin_value = origin

        headers: list[tuple[bytes, bytes]] = [
            (b"access-control-allow-origin", allow_origin_value.encode("latin1")),
            (b"vary", b"Origin"),
            (b"access-control-allow-methods", ", ".join(allow_methods).encode("latin1")),
            (b"access-control-allow-headers", ", ".join(allow_headers).encode("latin1")),
        ]
        if opts["allow_credentials"]:
            headers.append((b"access-control-allow-credentials", b"true"))
        if opts.get("max_age") is not None:
            headers.append((b"access-control-max-age", str(int(opts["max_age"])).encode("latin1")))
        return headers

    def _build_cors_headers_for_response(
        self,
        *,
        origin: Optional[str],
        opts: dict[str, Any]
    ) -> list[tuple[bytes, bytes]]:
        if not origin:
            return []

        # If "*" configured but credentials allowed: echo Origin instead.
        allow_origin_value = "*"
        if opts["allow_credentials"] or ("*" not in opts["allow_origins"]):
            allow_origin_value = origin

        headers: list[tuple[bytes, bytes]] = [
            (b"access-control-allow-origin", allow_origin_value.encode("latin1")),
            (b"vary", b"Origin"),
        ]
        if opts["allow_credentials"]:
            headers.append((b"access-control-allow-credentials", b"true"))
        expose = opts.get("expose_headers") or []
        if expose:
            headers.append((b"access-control-expose-headers", ", ".join(expose).encode("latin1")))
        return headers
    
    async def _cors_forbidden(self, send):
        await send({
            "type": "http.response.start",
            "status": 403,
            "headers": [(b"content-type", b"application/json")],
        })
        await send({
            "type": "http.response.body",
            "body": b'{"status":"error","message":"CORS origin not allowed"}',
        })

    async def _method_not_allowed(self, send, message: bytes = b"Preflight method not allowed"):
        await send({
            "type": "http.response.start",
            "status": 405,
            "headers": [(b"content-type", b"text/plain; charset=utf-8")],
        })
        await send({
            "type": "http.response.body",
            "body": message,
        })

    async def _handle_preflight(self, scope, send, origin: str, cors_opts: dict):
        acr_method = self._get_header(scope, "access-control-request-method")
        acr_headers = self._get_header(scope, "access-control-request-headers")

        if acr_method and cors_opts["allow_methods"] and acr_method.upper() not in cors_opts["allow_methods"]:
            return await self._method_not_allowed(send)

        headers = self._build_cors_headers_for_preflight(
            origin=origin,
            request_method=acr_method,
            request_headers=acr_headers,
            opts=cors_opts,
        )
        await send({
            "type": "http.response.start",
            "status": 204,  # No Content
            "headers": headers,
        })
        await send({"type": "http.response.body", "body": b""})

    def _wrap_cors_send(self, send, origin: str, cors_opts: dict):
        cors_headers = self._build_cors_headers_for_response(origin=origin, opts=cors_opts)
        async def cors_send(event):
            if event.get("type") == "http.response.start":
                headers = event.get("headers", [])
                event = {**event, "headers": headers + cors_headers}
            await send(event)
        return cors_send



# base_protocol.py
#pylint: disable=W0613

from __future__ import annotations

from typing import TYPE_CHECKING
from sqlalchemy import MetaData
from sqlalchemy.ext.asyncio import AsyncSession

if TYPE_CHECKING:
    from .sqlalchemy_models import AsyncQuery

class BaseModel:
    """
    Defines the interface that the custom
    DeclarativeBase class must satisfy.
    """

    metadata: MetaData

    @classmethod
    def query(cls, session: AsyncSession) -> "AsyncQuery": # type: ignore[empty-body]
        pass

    @classmethod
    def db_name(cls) -> str: # type: ignore[empty-body]
        pass



"""Controller class for endpoint groups"""


from typing import (Callable, TYPE_CHECKING, Optional, Type, TypeVar)

from pydantic import BaseModel
from ..media_types import MediaType
from ..http_statuses import HttpStatus

if TYPE_CHECKING:
    from ..pyjolt import PyJolt

T = TypeVar("T", bound="Controller")

def path(url_path: str = "/", open_api_spec: bool = True, tags: Optional[list[str]] = None) -> Callable[[Type[T]], Type[T]]:
    def decorator(cls: Type[T]) -> Type[T]:
        setattr(cls, "_controller_path", url_path)
        setattr(cls, "_include_open_api_spec", open_api_spec)
        setattr(cls, "_open_api_tags", tags)
        return cls

    return decorator

class Controller:

    def __init__(self, app: "PyJolt", path: str = "/", open_api_spec: bool = True, open_api_tags: Optional[list[str]] = None):
        self._app = app
        self._path = path
        self._before_request_methods: list[Callable] = []
        self._after_request_methods: list[Callable] = []
        self._controller_decorator_methods: list[Callable] = []
        self._endpoints_map: dict[str, dict[str, str|Callable]] = {}
        self._open_api_spec = open_api_spec
        self._open_api_tags = open_api_tags if open_api_tags is not None else [self.__class__.__name__]
        self.get_before_request_methods()
        self.get_after_request_methods()


    def get_endpoint_methods(self) -> dict[str, dict[str, str|Callable]]:
        """Returns a dictionery with all endpoint methods"""
        owner_cls: "type[Controller]|None" = self.__class__ or None
        endpoints: dict[str, dict] = {
            "GET": {},
            "POST": {},
            "PUT": {},
            "PATCH": {},
            "DELETE": {}
        }
        if owner_cls is None:
            return endpoints

        for name in dir(self):
            method = getattr(self, name)
            if not callable(method):
                continue
            endpoint_handler = getattr(method, "_handler", None)
            if endpoint_handler:
                endpoint_handler["tags"].extend(self._open_api_tags)
                http_method: str = endpoint_handler.get("http_method") # type: ignore
                endpoints[http_method][endpoint_handler["path"]] = {"method": method,
                                                                    "base_path": self._path,
                                                                    **endpoint_handler}
        self._endpoints_map = endpoints
        return endpoints
    
    def get_before_request_methods(self):
        owner_cls = self.__class__ or None
        if owner_cls is None:
            return

        for name in dir(self):
            method = getattr(self, name)
            if not callable(method):
                continue
            handler = getattr(method, "_before_request", None)
            if handler:
              self._before_request_methods.append(method)  
    
    def get_after_request_methods(self):
        owner_cls = self.__class__ or None
        if owner_cls is None:
            return

        for name in dir(self):
            method = getattr(self, name)
            if not callable(method):
                continue
            handler = getattr(method, "_after_request", None)
            if handler:
              self._after_request_methods.append(method)

    @property
    def endpoints_map(self) -> dict[str, dict[str, str|Callable]]:
        """Returns map of all endpoints"""
        return self._endpoints_map

    @property
    def path(self) -> str:
        """Path variable of the class"""
        return self._path
    
    @property
    def open_api_spec(self) -> bool:
        """If it is included in the open api specs"""
        return self._open_api_spec

    @property
    def app(self) -> "PyJolt":
        """App object"""
        return self._app


class Descriptor:

    def __init__(self, status: int|HttpStatus = HttpStatus.BAD_REQUEST,
                 description: Optional[str] = None,
                 media_type: MediaType = MediaType.APPLICATION_JSON,
                 body: Optional[Type[BaseModel]] = None):
        if isinstance(status, HttpStatus):
            status = status.value
        self._status = status
        self._description = description
        self._body = body
        self._media_type = media_type

    @property
    def status(self) -> HttpStatus|int:
        return self._status

    @property
    def description(self) -> Optional[str]:
        return self._description

    @property
    def body(self) -> Optional[Type[BaseModel]]:
        return self._body

    @property
    def media_type(self) -> MediaType:
        return self._media_type



"""
migrate.py
Alembic integration for database migrations, with argparse subparsers for CLI commands.
"""

import os
import shutil
from typing import Optional, cast
from alembic.config import Config
from alembic import command
from sqlalchemy.orm import declarative_base

from ...pyjolt import PyJolt
from ..sql_database import SqlDatabase
from ...base_extension import BaseExtension

def register_db_commands(app: PyJolt, migrate: 'Migrate'):
    """
    Registers subparsers (CLI commands) for database migrations to the app's CLI.
    This replaces the old app.add_cli_command(...) calls with subparser definitions
    so that we can parse arguments like "db-upgrade <revision>" or "db-migrate --message '...'",
    etc.
    """

    # db-init
    sp_init = app.subparsers.add_parser(f"{migrate.command_prefix}db-init", help="Initialize the Alembic migration environment.")
    sp_init.set_defaults(func=lambda args: migrate.init())

    # db-migrate
    sp_migrate = app.subparsers.add_parser(f"{migrate.command_prefix}db-migrate", help="Generate a new revision file (autogenerate).")
    sp_migrate.add_argument("--message", default="", help="Revision message.")
    sp_migrate.set_defaults(func=lambda args: migrate.migrate(message=args.message))

    # db-upgrade
    sp_upgrade = app.subparsers.add_parser(f"{migrate.command_prefix}db-upgrade", help="Upgrade the database to a specified (or head) revision.")
    sp_upgrade.add_argument("--revision", nargs="?", default="head", help="Revision identifier (default=head).")
    sp_upgrade.set_defaults(func=lambda args: migrate.upgrade(revision=args.revision))

    # db-downgrade
    sp_downgrade = app.subparsers.add_parser(f"{migrate.command_prefix}db-downgrade", help="Downgrade the database to a specified (or one step) revision.")
    sp_downgrade.add_argument("--revision", nargs="?", default="-1", help="Revision identifier (default=-1, i.e. one step down).")
    sp_downgrade.set_defaults(func=lambda args: migrate.downgrade(revision=args.revision))

    # db-history
    sp_history = app.subparsers.add_parser(f"{migrate.command_prefix}db-history", help="Show revision history.")
    sp_history.add_argument("--verbose", action="store_true", help="Show more details about each revision.")
    sp_history.add_argument("--indicate-current", action="store_true", help="Mark the current revision in the log.")
    sp_history.set_defaults(func=lambda args: migrate.history(verbose=args.verbose, indicate_current=args.indicate_current))

    # db-current
    sp_current = app.subparsers.add_parser(f"{migrate.command_prefix}db-current", help="Show the current revision of the database.")
    sp_current.add_argument("--verbose", action="store_true", help="Show more details about the current revision.")
    sp_current.set_defaults(func=lambda args: migrate.current(verbose=args.verbose))

    # db-heads
    sp_heads = app.subparsers.add_parser(f"{migrate.command_prefix}db-heads", help="Show all current 'head' revisions.")
    sp_heads.add_argument("--verbose", action="store_true", help="Show more details.")
    sp_heads.set_defaults(func=lambda args: migrate.heads(verbose=args.verbose))

    # db-show
    sp_show = app.subparsers.add_parser(f"{migrate.command_prefix}db-show", help="Show details of a given revision.")
    sp_show.add_argument("--revision", nargs="?", default="head", help="The revision to show (default=head).")
    sp_show.set_defaults(func=lambda args: migrate.show(revision=args.revision))

    # db-stamp
    sp_stamp = app.subparsers.add_parser(f"{migrate.command_prefix}db-stamp", help="Stamp the database with a given revision (no actual migration).")
    sp_stamp.add_argument("--revision", nargs="?", default="head", help="Revision to stamp (default=head).")
    sp_stamp.set_defaults(func=lambda args: migrate.stamp(revision=args.revision))


class Migrate(BaseExtension):
    """
    Integrates Alembic with the application for managing database migrations.
    Uses the same variables prefix as the SqlDatabase instance

    The command prefix is used to differentiate between different Migration instances
    when using the CLI.
    """
    def __init__(self, db: SqlDatabase, command_prefix: str = ""):
        self._app: "PyJolt"
        self._root_path: str
        self._db: SqlDatabase = db
        self._variable_prefix: Optional[str] = None
        self._migrations_path: Optional[str] = None
        self._migration_dir: Optional[str] = None
        self._database_uri: Optional[str] = None
        self._command_prefix: str = command_prefix

    def init_app(self, app: PyJolt):
        """
        Initializes Alembic for the given app.
        """
        self._app = app
        self._root_path = self._app.root_path
        self._variable_prefix = self._db.variable_prefix
        self._migration_dir = self._app.get_conf(f"{self._variable_prefix}ALEMBIC_MIGRATION_DIR", "migrations")
        self._migrations_path = os.path.join(self._root_path, cast(str, self._migration_dir))
        # use a SYNC database URI for migrations
        self._database_uri = self._app.get_conf(f"{self._variable_prefix}ALEMBIC_DATABASE_URI_SYNC")
        app.add_extension(self)
        # Register all the subparser commands
        register_db_commands(self._app, self)

    def get_alembic_config(self) -> Config:
        """
        Returns an Alembic configuration object.
        """
        cfg_path = os.path.join(cast(str, self._migrations_path), "alembic.ini")
        config = Config(cfg_path)
        config.set_main_option("sqlalchemy.url", cast(str, self._database_uri))
        #pylint: disable-next=W0212
        associated_models = self._app._db_models.get(self._db.db_name, None)
        if not associated_models or len(associated_models) == 0:
            #pylint: disable-next=W0719
            raise Exception(f"No models associated with db: {self._db.db_name}")
        config.attributes["target_metadata"] = associated_models[0].metadata
        return config

    def _copy_env_template(self):
        """
        Copies the env.py template to the migrations directory.
        """
        template_path = os.path.join(os.path.dirname(__file__), "env_template.py")
        destination_path = os.path.join(self._migrations_path, "env.py")
        shutil.copy(template_path, destination_path)

    def init(self):
        """
        Initializes the Alembic migration environment (if not done already).
        """
        if not os.path.exists(self._migrations_path):
            os.makedirs(self._migrations_path)
        command.init(self.get_alembic_config(), self._migrations_path, template="generic")
        self._copy_env_template()

    def migrate(self, message="Generate migration"):
        """
        Creates a new Alembic revision script (autogenerated) with the given message.
        """
        config = self.get_alembic_config()
        command.revision(config, autogenerate=True, message=message)

    def upgrade(self, revision="head"):
        """
        Upgrades the database to the given revision (default=head).
        """
        config = self.get_alembic_config()
        command.upgrade(config, revision)

    def downgrade(self, revision="-1"):
        """
        Downgrades the database to the given revision (default=-1, i.e. one step).
        """
        config = self.get_alembic_config()
        command.downgrade(config, revision)

    def history(self, verbose=False, indicate_current=False):
        """
        Shows the revision history.
        """
        config = self.get_alembic_config()
        command.history(config, verbose=verbose, indicate_current=indicate_current)

    def current(self, verbose=False):
        """
        Shows the current revision of the database.
        """
        config = self.get_alembic_config()
        command.current(config, verbose=verbose)

    def stamp(self, revision="head"):
        """
        Stamps the database with a given revision, without running migrations.
        """
        config = self.get_alembic_config()
        command.stamp(config, revision)

    def heads(self, verbose=False):
        """
        Displays all current head revisions.
        """
        config = self.get_alembic_config()
        command.heads(config, verbose=verbose)

    def show(self, revision="head"):
        """
        Shows details of a specific revision (default=head).
        """
        config = self.get_alembic_config()
        command.show(config, revision)

    @property
    def command_prefix(self) -> str:
        """
        Returns the command prefix for the CLI
        """
        return self._command_prefix



"""
Base configuration class
"""

from __future__ import annotations

import re
from typing import Optional, List
from pydantic import BaseModel, Field, ConfigDict, field_validator

IMPORT_STR_RE = re.compile(r"^[A-Za-z_]\w*(?:\.[A-Za-z_]\w*)*:[A-Za-z_]\w*$")


class BaseConfig(BaseModel):
    model_config = ConfigDict(extra="allow")

    # required
    APP_NAME: str = Field(description="Human-readable name of the app")
    VERSION: str = Field(description="Application version")
    BASE_PATH: str = Field(
        description="Base path of app. os.path.dirname(__file__) in the configs.py is the usual value."
    )

    # required for Authentication extension
    SECRET_KEY: Optional[str] = Field(
        None, description="High entropy random string for signing cookies/jwts"
    )

    # optionals with sensible defaults
    DEBUG: Optional[bool] = Field(
        True, description="If the app should run in debug mode or not."
    )
    HOST: Optional[str] = Field(
        "localhost", description="Host ip where the app should run."
    )
    PORT: Optional[int] = Field(8080, description="Port on which the app should run.")
    LIFESPAN: Optional[str] = Field("on", description="Whether to use server lifespan events/signals. Options: on, auto, off")
    TEMPLATES_DIR: Optional[str] = Field(
        "/templates", description="Relative templates dir from root"
    )
    STATIC_DIR: Optional[str] = Field(
        "/static", description="Relative static dir from root"
    )
    STATIC_URL: Optional[str] = Field(
        "/static", description="URL prefix for static files"
    )
    STATIC_CONTROLLER_NAME: Optional[str] = Field("static", description="Mount name for static files controller")
    TEMPLATES_STRICT: Optional[bool] = Field(
        True, description="Strict template rendering"
    )
    STRICT_SLASHES: Optional[bool] = Field(
        False, description="Route '/x' vs '/x/' strictness"
    )
    OPEN_API: Optional[bool] = Field(True, description="Enable OpenAPI endpoint")
    OPEN_API_URL: Optional[str] = Field("/openapi", description="OpenAPI base path")
    OPEN_API_DESCRIPTION: Optional[str] = Field(
        "Simple API", description="OpenAPI description"
    )

    #CORS settings
    CORS_ENABLED: Optional[bool] = Field(True, description="Enable CORS")
    CORS_ALLOW_ORIGINS: Optional[list[str]] = Field(["*"], description="List of allowed origins")
    CORS_ALLOW_METHODS: Optional[list[str]] = ["GET", "POST", "PUT", "PATCH", "DELETE", "OPTIONS"]
    CORS_ALLOW_HEADERS: Optional[list[str]] = Field(["Authorization", "Content-Type"], description="List of allowed headers")
    CORS_EXPOSE_HEADERS: Optional[list[str]] = Field([], description="Expose headers")
    CORS_ALLOW_CREDENTIALS: Optional[bool] = Field(True, description="Allow credentials")
    CORS_MAX_AGE: Optional[int] = Field(None, description="Max age in seconds. None to disable.")

    # controllers, cli_controllers, extensions, models, exception handlers and middleware to load
    CONTROLLERS: Optional[List[str]] = None
    CLI_CONTROLLERS: Optional[List[str]] = None
    EXTENSIONS: Optional[List[str]] = None
    MODELS: Optional[List[str]] = None
    EXCEPTION_HANDLERS: Optional[List[str]] = None
    MIDDLEWARE: Optional[List[str]] = None

    @field_validator("CONTROLLERS", "CLI_CONTROLLERS", "EXTENSIONS", "MODELS", "EXCEPTION_HANDLERS", "MIDDLEWARE", mode="before")
    @classmethod
    def _coerce_list_of_str(cls, v):
        if v is None:
            return None
        if not isinstance(v, list) or any(not isinstance(x, str) for x in v):
            raise TypeError("Must be a list[str] or None.")
        return v

    @field_validator("CONTROLLERS", "CLI_CONTROLLERS", "EXTENSIONS", "MODELS", "EXCEPTION_HANDLERS", "MIDDLEWARE")
    @classmethod
    def _validate_import_strings(cls, v):
        if not v:
            return v
        bad = [s for s in v if not IMPORT_STR_RE.match(s)]
        if bad:
            raise ValueError(
                "Invalid import strings (use 'package.module:ClassOrObject'): "
                + ", ".join(bad)
            )
        return v
    
    @staticmethod
    def value_to_bool(value: str|int|bool) -> bool:
        """
        Turns a boolean-like value to boolean.

        :param str value: a string value representing a boolean

        Returns True if value in [True, "true", "True", "1", 1]
        """
        return value in [True, "true", "True", "1", 1]



"""
Users API
"""
from app.api.models import Role, User
from app.api.users_api.dtos import ErrorResponse, TestModel, TestModelOut, TestModelOutList
from app.authentication import UserRoles
from app.extensions import db, cache

import asyncio
from pyjolt import HttpStatus, MediaType, Request, Response, html_abort
from pyjolt.controller import (
    Controller,
    Descriptor,
    consumes,
    delete,
    get,
    open_api_docs,
    path,
    post,
    produces
)
from pyjolt.database import AsyncSession


@path("/api/v1/users", tags=["Users"])
class UsersApi(Controller):

    @get("/")
    @produces(MediaType.APPLICATION_JSON)
    @db.managed_session
    @cache.cache(duration=5)
    async def get_users(self, req: Request, session: AsyncSession) -> Response[TestModelOutList]:
        """Endpoint for returning all app users"""
        #await asyncio.sleep(10)
        await asyncio.sleep(10)
        users = await User.query(session).all()
        response = {
            "message": "Users fetched successfully",
            "status": "success",
            "data": [TestModel(fullname=user.fullname, age=user.age, email=user.email) for user in users] if users else None
        }
        return req.response.json(response).status(HttpStatus.OK)

    @get("/<int:user_id>")
    @produces(MediaType.APPLICATION_JSON)
    @open_api_docs(
        Descriptor(status=HttpStatus.NOT_FOUND, description="User not found", body=ErrorResponse),
        Descriptor(status=HttpStatus.BAD_REQUEST, description="Bad request", body=ErrorResponse))
    @cache.cache(duration=20)
    async def get_user(self, req: Request, user_id: int) -> Response[TestModelOut]:
        """Returns single user by id"""
        if user_id > 10:
            return html_abort("index.html", HttpStatus.CONFLICT)
        await asyncio.sleep(10)
        return req.response.json({
            "message": "User fetched successfully",
            "status": "success",
            "data": TestModel(fullname="Marko Šterk", age=22, email="marko_sterk@hotmail.com")
        }).status(HttpStatus.OK)

    @post("/")
    @consumes(MediaType.APPLICATION_JSON)
    @produces(MediaType.APPLICATION_JSON)
    @db.managed_session
    async def create_user(self, req: Request, data: TestModel, session: AsyncSession) -> Response[TestModelOut]:
        """Consumes and produces json"""
        user: User = await User.query(session).filter_by(email=data.email).first()
        if user:
            return req.response.json({
                "message": "User with this email already exists",
                "status": "error"
            }).status(HttpStatus.BAD_REQUEST)
        user = User(email=data.email, fullname=data.fullname, age=data.age)
        session.add(user)
        await session.flush()
        role = Role(user_id=user.id, role=UserRoles.ADMIN)
        session.add(role)
        return req.response.json({
            "message": "User added successfully",
            "status": "success"
        }).status(200)

    @delete("/<int:user_id>")
    @produces(media_type=MediaType.NO_CONTENT, status_code=HttpStatus.NO_CONTENT)
    @open_api_docs(
        Descriptor(status=HttpStatus.NOT_FOUND, description="User not found", body=ErrorResponse
    ))
    @db.managed_session
    async def delete_user(self, req: Request, user_id: int, session: AsyncSession) -> Response:
        """Deletes user"""
        user: User = await User.query(session).filter_by(id=user_id).first()
        if not user:
            return req.response.json({
                "message": "User with this id does not exist",
                "status": "error"
            }).status(HttpStatus.NOT_FOUND)

        await session.delete(user)
        return req.response.no_content()



"""
Base extension class
"""
from typing import TYPE_CHECKING
from abc import abstractmethod

if TYPE_CHECKING:
    from .pyjolt import PyJolt

class BaseExtension:
    
    @abstractmethod
    def init_app(self, app: "PyJolt") -> None:
        ...



"""
Chat session model for AI interface
"""
from sqlalchemy import ForeignKey
from sqlalchemy.orm import mapped_column, Mapped

from .base_model import BaseModel

class ChatSession(BaseModel):

    __tablename__ = "chat_sessions"

    user_id: Mapped[int] = mapped_column(
        ForeignKey("users.id", ondelete="CASCADE"),
        nullable=True,
        index=True
    )



"""
App configurations
"""
import os
from typing import cast #,Type
from pyjolt import BaseConfig
#from pyjolt.caching.memory_cache_backend import MemoryCacheBackend

class Config(BaseConfig):
    """Config class"""
    APP_NAME: str = cast(str, os.environ.get("APP_NAME"))
    VERSION: str = cast(str, os.environ.get("VERSION"))
    SECRET_KEY: str = cast(str, os.environ.get("SECRET_KEY"))
    BASE_PATH: str = os.path.dirname(__file__)
    DEBUG: bool = BaseConfig.value_to_bool(os.environ.get("DEBUG", "True"))

    DATABASE_URI: str = cast(str, os.environ.get("DATABASE_URI"))
    ALEMBIC_DATABASE_URI_SYNC: str = cast(str, os.environ.get("ALEMBIC_DATABASE_URI_SYNC"))

    #CACHE_BACKEND: Type[MemoryCacheBackend] = MemoryCacheBackend #default is MemoryCacheBackend
    #CACHE_DURATION: int = 500 #default is 300 s

    CONTROLLERS: list[str] = [
        'app.api.auth_api:AuthApi',
        'app.api.users_api.users_api:UsersApi',
        'app.page.page_controller:PageController'
    ]

    CLI_CONTROLLERS: list[str] = [
        'app.cli.cli_controller:UtilityCLIController'
    ]

    EXTENSIONS: list[str] = [
        'app.extensions:db',
        'app.extensions:migrate',
        'app.extensions:cache',
        'app.authentication:auth',
        'app.ai_interface:ai_interface'
    ]

    MODELS: list[str] = [
        'app.api.models:User',
        'app.api.models:Role'
    ]

    EXCEPTION_HANDLERS: list[str] = [
        'app.api.exceptions.exception_handler:CustomExceptionHandler'
    ]

    MIDDLEWARE: list[str] = [
        'app.middleware.timing_mw:TimingMW'
    ]



"""
Authentication
"""

from typing import Any
from pyjolt.auth import Authentication
from pyjolt import Request

class Auth(Authentication):
    """This class must implement the user_loader and role_check (optional) methods"""

    async def user_loader(self, req: Request) -> Any:
        """
        Implement this method for loading users from the request object.
        Method should return the user object (loaded from db or the jwt).

        return None -> user was not found/loaded. User is not authenticated
        return Any -> user is authenticated
        """
        return None

    async def role_check(self, user: Any, roles: list[Any]) -> bool:
        """
        Implement check if user has required roles.

        True -> user has role
        False -> user doesn't have role
        """
        return False

auth: Auth = Auth()



"""
Base model classes for SQLAlchemy models.
"""
import logging
from typing import Any, Type, TypeVar, Dict
from sqlalchemy import func
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.future import select
from sqlalchemy.orm import declarative_base
from sqlalchemy.sql import Select

from .base_protocol import BaseModel

def create_declerative_base(name: str = "db") -> Type[BaseModel]:
    """
    Declarative base class factory that returns a type
    satisfying the BaseModel.

    :param str name: The name of the associated database. The argument passed when creating the db extension instance.
    """
    base = declarative_base()

    class DeclarativeBase(base, BaseModel):  # type: ignore
        """
        Base model from sqlalchemy.orm with
        query classmethod
        """
        __abstract__ = True

        __db_name__: str = name

        @classmethod
        def query(cls, session: AsyncSession) -> "AsyncQuery":
            """
            Creates an AsyncQuery instance
            """
            return AsyncQuery(session, cls)

        @classmethod
        def db_name(cls) -> str:
            return cls.__db_name__

    return DeclarativeBase

#pylint: disable-next=C0103
_T0 = TypeVar("_T0", bound=Any)

class AsyncQuery:
    """
    Async-friendly intuitive query object.
    Easy and intuitive querying with pagination support.
    """

    def __init__(self, session: AsyncSession, model: Type[_T0]):
        self.session = session
        self.model = model
        self._query: Select = select(model)  # Start with SELECT * FROM table

    def where(self, *conditions) -> "AsyncQuery":
        """Adds WHERE conditions (same as `filter()`)."""
        return self.filter(*conditions)

    def filter(self, *conditions) -> "AsyncQuery":
        """Adds WHERE conditions to the query (supports multiple conditions)."""
        self._query = self._query.filter(*conditions)
        return self

    def filter_by(self, **kwargs) -> "AsyncQuery":
        """Adds WHERE conditions using keyword arguments (simpler syntax)."""
        self._query = self._query.filter_by(**kwargs)
        return self

    def join(self, other_model: Type[_T0]) -> "AsyncQuery":
        """Performs a SQL JOIN with another model."""
        self._query = self._query.join(other_model)
        return self

    def limit(self, num: int) -> "AsyncQuery":
        """Limits the number of results returned."""
        self._query = self._query.limit(num)
        return self

    def offset(self, num: int) -> "AsyncQuery":
        """Skips a certain number of results (used for pagination)."""
        self._query = self._query.offset(num)
        return self

    def order_by(self, *columns) -> "AsyncQuery":
        """Sorts results based on one or more columns."""
        self._query = self._query.order_by(*columns)
        return self
    
    def like(self, column, pattern, escape=None) -> "AsyncQuery":
        """
        Filters results using a SQL LIKE condition.

        :param column: A column attribute, e.g. User.name
        :param pattern: The pattern string (with % wildcards)
        :param escape: The escape character, if needed.
        :return: self, for chaining
        """
        self._query = self._query.filter(column.like(pattern, escape=escape))
        return self

    def ilike(self, column, pattern, escape=None) -> "AsyncQuery":
        """
        Filters results using a case-insensitive LIKE (ILIKE) condition.
        (Supported by Postgres, some other dialects).

        :param column: A column attribute, e.g. User.name
        :param pattern: The pattern string (with % wildcards)
        :param escape: The escape character, if needed.
        :return: self, for chaining
        """
        self._query = self._query.filter(column.ilike(pattern, escape=escape))
        return self

    async def count(self) -> int:
        """
        Returns the total number of records matching the current query,
        preserving any applied filters.
        """
        count_query = select(func.count()).select_from(self.model)
        # Apply existing filters
        if hasattr(self._query, "whereclause") and self._query.whereclause is not None:
            count_query = count_query.where(self._query.whereclause)
        self._query = count_query
        result = await self._execute_query()
        result: int = result.scalar() or 0
        return result

    async def paginate(self, page: int = 1, per_page: int = 10) -> Dict[str, Any]:
        """
        Paginates results.
        page (int): The page number (1-based index).
        per_page (int): Number of results per page.
        ```
        Returns:
            dict: {
                "items": List of results,
                "total": Total records,
                "page": Current page,
                "pages": Total pages,
                "per_page": Results per page,
                "has_next": Whether there's a next page,
                "has_prev": Whether there's a previous page
            }
        ```
        """
        page = max(page, 1)

        total_records = await self.count()
        total_pages = (total_records + per_page - 1) // per_page  # Round up division

        self._query = self._query.limit(per_page).offset((page - 1) * per_page)
        result = await self._execute_query()
        items = result.scalars().all()
        #await self.session.close()

        return {
            "items": items,
            "total": total_records,
            "page": page,
            "pages": total_pages,
            "per_page": per_page,
            "has_next": page < total_pages,
            "has_prev": page > 1,
        }

    async def _execute_query(self) -> Any:
        """Executes the query safely with automatic rollback on failure."""
        try:
            result = await self.session.execute(self._query)
            return result
        except SQLAlchemyError as e:
            await self.session.rollback()
            logging.error("Database query failed: %s", e)
            raise

    async def all(self) -> list:
        """Executes the query and returns all results."""
        result = await self._execute_query()
        return result.scalars().all()

    async def first(self) -> Any:
        """Executes the query and returns the first result."""
        result = await self._execute_query()
        return result.scalars().first()

    async def one(self) -> Any:
        """Executes the query and expects exactly one result."""
        result = await self._execute_query()
        return result.scalars().one()




"""
Test setup
"""

import pytest
from pyjolt.testing import PyJoltTestClient
from app import Application

@pytest.fixture
async def application():
    yield Application()

@pytest.fixture
async def client(application):
    async with PyJoltTestClient(application) as c:
        yield c



"""
Helpers for creating embeddings
"""
import numpy as np
from torch import Tensor
from sentence_transformers import SentenceTransformer
from pgvector.sqlalchemy import Vector as VectorColumn

__all__ = ['l2_distance', 'cosine_similarity', 'cosine_distance',
            'create_embedding','chunkify_text', 'VectorColumn']

def l2_distance(vec1: list[float]|np.ndarray,
                vec2: list[float]|np.ndarray) -> float:
    """
    Calculates l2 distance between two vectors

    :param vec1: first vector.
    :param vec2: second vector.
    """
    if isinstance(vec1, list):
        vec1 = np.array(vec1)
    if isinstance(vec2, list):
        vec2 = np.array(vec2)
    return np.sqrt(np.sum((np.array(vec1) - np.array(vec2)) ** 2))


def cosine_similarity(vec1: list[float]|np.ndarray,
                vec2: list[float]|np.ndarray) -> float:
    """
    Calculates cosine similarity between two vectors

    :param vec1: first vector.
    :param vec2: second vector.
    """
    if isinstance(vec1, list):
        vec1 = np.array(vec1)
    if isinstance(vec2, list):
        vec2 = np.array(vec2)
    dot_product = np.dot(vec1, vec2)
    norm_vec1 = np.linalg.norm(vec1)
    norm_vec2 = np.linalg.norm(vec2)
    return dot_product / (norm_vec1 * norm_vec2)

def cosine_distance(vec1: list[float]|np.ndarray,
                vec2: list[float]|np.ndarray) -> float:
    """
    Calculates cosine distance between two vectors
    
    :param vec1: first vector.
    :param vec2: second vector.
    :returns: cosine distance as 1 - cosimn_similarity(vec1, vec2)
    """
    similarity: float = cosine_similarity(vec1, vec2)
    return 1 - similarity  # Cosine distance is 1 - cosine similarity

def create_embedding(text: str,
                     transformer: str = "infgrad/stella-base-en-v2",
                     trust_remote_code: bool = True,
                     **kwargs) -> Tensor:
    """
    Creates embedding for provided text with specified transformer

    :param text: text for which the embedding is created.
    :param transformer: sentence transfor that is used.
    :param trust_remote_code: if you allow code from the Hugging Face Hun repo to be executed locally
    :param kwargs: any keyword arguments that are accepted by the SentenceTransformer class
    """
    embedding_model = SentenceTransformer(transformer,
                                          trust_remote_code=trust_remote_code,
                                          **kwargs)
    return embedding_model.encode(text)

def chunkify_text(text: str, chunk_size: int) -> list[str]:
    """
    Takes a text and creates a list of chunks with words <= chunk_size

    :param text: the text to be chunkified.
    :param chunk_size: max number of words in chunk
    """
    words: list[str] = text.split()
    chunks: list[str] = []
    current_chunk: list[str] = []

    for word in words:
        current_chunk.append(word)
        if len(current_chunk) == chunk_size:
            chunks.append(' '.join(current_chunk))
            current_chunk = []
    if len(current_chunk) > 0:
        chunks.append(' '.join(current_chunk))
    return chunks



from pyjolt.cli import CLIController, command, argument

class UtilityCLIController(CLIController):
    """A simple CLI utility controller."""

    @command("greet", help="Greet a user with a message.")
    @argument("name", arg_type=str, description="The name of the user to greet.")
    async def greet(self, name: str):
        """Greet by name."""
        print(f"Hello, {name}! Welcome to the CLI utility.")

    @command("add", help="Add two numbers.")
    @argument("a", arg_type=int, description="The first number.")
    @argument("b", arg_type=int, description="The second number.")
    async def add(self, a: int, b: int):
        """Add two numbers and print the result."""
        result = a + b
        print(f"The sum of {a} and {b} is {result}.")
    
    @command("goodbye", help="Says goodbye")
    async def say_goodbye(self):
        """Say goodbye when the CLI exits."""
        print("Goodbye! Thanks for using the CLI utility.")



"""
Example data model
"""
from sqlalchemy import String
from sqlalchemy.orm import mapped_column, Mapped

from pyjolt.database import create_declerative_base

Base = create_declerative_base()

class Example(Base):
    """
    Example model
    """
    #table name in database; usually plural
    __tablename__: str = "examples"

    id: Mapped[int] = mapped_column(primary_key=True)
    name: Mapped[str] = mapped_column(String(30))



"""
Collection of http exceptions that can be raised
"""
from typing import Any, Optional
from pydantic import ValidationError as PydanticValidationError
from ..http_statuses import HttpStatus
from ..http_statuses import HttpStatus

class BaseHttpException(Exception):
    """
    Base http exception class
    """
    def __init__(self, message: str = "",
                 status_code: int|HttpStatus= 500,
                 status = "error",
                 data = None):
        """
        Init method
        """
        self.message = message
        if isinstance(status_code, HttpStatus):
            status_code = status_code.value
        self.status_code = status_code
        self.status = status
        self.data = data

class StaticAssetNotFound(BaseHttpException):
    """
    HTTP exception for static assets not found
    """
    def __init__(self, message: str = "Static asset not found",
                 status_code: int = 404,
                 status: str = "error",
                 data: Any = None):
        super().__init__(
            message,
            status_code,
            status,
            data
        )


class AborterException(BaseHttpException):
    """
    Aborter exception
    """
    def __init__(self, message: str = "",
                 status_code: int = 400,
                 status: str = "error",
                 data: Optional[Any] = None):
        super().__init__(
            message,
            status_code,
            status,
            data
        )

class HtmlAborterException(BaseHttpException):
    """
    Html aborter exception
    """
    def __init__(self, template: str, status_code: int|HttpStatus, data: Optional[Any] = None):
        super().__init__("Error", status_code=status_code, data=data)
        self.template = template
        if isinstance(status_code, HttpStatus):
            status_code = status_code.value
        self.status_code = status_code
        self.data = data

class MissingRequestData(BaseHttpException):
    """
    Exception for missing request data. 
    Raised by input decorator on route handlers
    """
    def __init__(self, message: str = "",
                 status_code: int = 400,
                 status: str = "error",
                 data: Any = None):
        super().__init__(
            message,
            status_code,
            status,
            data
        )

class SchemaValidationError(BaseHttpException, PydanticValidationError):
    """
    Exception for schema validation errors
    """
    def __init__(self, message: list[str]|list|dict):
        super().__init__(
            "Data validation failed",
            422,
            "error",
            message
        )

class PydanticSchemaValidationError(BaseHttpException):
    """
    Exception for schema validation errors with Pydantic
    """
    def __init__(self, messages: list[dict]):
        parsed_messages: dict[str, list[str]] = {}
        for obj in messages:
            loc: str = obj["loc"][0]
            if loc not in parsed_messages:
                parsed_messages[loc] = []
            parsed_messages[loc].append(obj["msg"])
        super().__init__("Data validation failed",
                         422,
                         "error",
                         parsed_messages)

class AuthenticationException(BaseHttpException):
    """
    Authentication exception for endpoints which require authentication
    """
    def __init__(self, message: str):
        super().__init__(
            message,
            HttpStatus.FORBIDDEN,
            "error",
            None
        )

class UnauthorizedException(BaseHttpException):
    """
    Authorization exception for endpoints which require specific roles
    """
    def __init__(self, message: str, roles: list[Any]):
        super().__init__(
            message,
            HttpStatus.UNAUTHORIZED,
            "error",
            roles
        )

class InvalidJWTError(BaseHttpException):
    """
    Invalid or expired JWT token error
    """
    def __init__(self, message: str):
        super().__init__(
            message,
            401,
            "error",
            None
        )

def abort(msg: str, status_code: int|HttpStatus=HttpStatus.BAD_REQUEST, status: str = "error", data: Any = None):
    """
    Aborts request by raising an aborter exception
    """
    if isinstance(status_code, HttpStatus):
        status_code = status_code.value
    raise AborterException(msg, status_code, status, data)

def html_abort(template: str, status_code: int|HttpStatus=HttpStatus.BAD_REQUEST, data: Any = None):
    """
    Aborts request with html response
    """
    if isinstance(status_code, HttpStatus):
        status_code = status_code.value
    raise HtmlAborterException(template, status_code, data)



"""
UsersAPI dtos
"""
from typing import Optional, Any
from pydantic import BaseModel, Field, field_serializer

class TestModel(BaseModel):
    fullname: str = Field(min_length=3, max_length=15)
    age: int = Field(gt=17)
    email: str = Field(min_length=5, max_length=30)

class BaseResponseModel(BaseModel):
    message: str
    status: str

class TestModelOut(BaseResponseModel):
    data: Optional[TestModel] = None

class TestModelOutList(BaseResponseModel):
    data: Optional[list[TestModel]] = None

class ErrorResponse(BaseResponseModel):
    data: Optional[Any] = None

    @field_serializer("data")
    def serialize_data(self, data: Any, _info):
        if isinstance(data, BaseModel):
            return data.model_dump()
        return data



"""
Chat API
"""
from pyjolt import Request, Response, HttpStatus, MediaType
from pyjolt.controller import Controller, get, path, produces

@path("/api/v1/chat")
class ChatApi(Controller):

    @get("/<int:chat_session_id>")
    @produces(MediaType.APPLICATION_JSON)
    async def get_chat_session(self, req: Request, chat_session_id: int) -> Response:

        return req.res.json({
            "chat_session_id": chat_session_id
        }).status(HttpStatus.OK)



"""
sql_database.py
Module for sql database connection/integration
"""

#import asyncio
from typing import Optional, Callable, cast, TYPE_CHECKING
from functools import wraps
from sqlalchemy.engine import RowMapping
from sqlalchemy.ext.asyncio import (
    AsyncEngine,
    create_async_engine,
    AsyncSession,
    async_sessionmaker
)
from pydantic import BaseModel, Field

from ..utilities import run_sync_or_async
from ..base_extension import BaseExtension

if TYPE_CHECKING:
    from ..pyjolt import PyJolt

class SqlDatabaseConfig(BaseModel):
    """Configuration options for SqlDatabase extension"""
    DATABASE_URI: str = Field(description="Connection string for the database")
    DATABASE_SESSION_NAME: str = Field("session", description="AsyncSession variable name for use with @managed_session decorator")

class SqlDatabase(BaseExtension):
    """
    A simple async Database interface using SQLAlchemy.
    """

    def __init__(self, db_name: str = "db", variable_prefix: str = "") -> None:
        self._app: "Optional[PyJolt]" = None
        self._engine: Optional[AsyncEngine] = None
        self._session_factory: Optional[async_sessionmaker[AsyncSession]] = None
        self._db_uri: Optional[str] = None
        self._variable_prefix: str = variable_prefix
        self.__db_name__ = db_name
        self.session_name: str = "session"

    def init_app(self, app: "PyJolt") -> None:
        """
        Initializes the database interface
        app.get_conf("DATABASE_URI") must return a connection string like:
        "postgresql+asyncpg://user:pass@localhost/dbname"
        or "sqlite+aiosqlite:///./test.db"
        """
        self._app = app
        self._db_uri = self._app.get_conf(f"{self._variable_prefix}DATABASE_URI")
        self.session_name = self._app.get_conf(f"{self._variable_prefix}DATABASE_SESSION_NAME",
                                               self.session_name)
        self._app.add_extension(self)
        self._app.add_on_startup_method(self.connect)
        self._app.add_on_shutdown_method(self.disconnect)

    async def connect(self) -> None:
        """
        Creates the async engine and session factory.
        Runs automatically when the lifespan.start signal is received
        """
        if not self._engine:
            self._engine = create_async_engine(
                cast(str, self._db_uri),
                echo=False,
                pool_pre_ping=True,
                pool_recycle=1800
            )

        self._session_factory = async_sessionmaker(
            bind=self._engine,
            expire_on_commit=False,
            autoflush=False
        )

    def create_session(self) -> AsyncSession:
        """
        Creates new session and returns session object. Used for manual session handling.

        WARNING: You must close the session manually after use with await session.close()
        or use it within an async with block.
        """
        if self._session_factory is not None:
            return cast(AsyncSession, self._session_factory())
        #pylint: disable-next=W0719
        raise Exception("Session factory is None")

    async def disconnect(self) -> None:
        """
        Runs automatically when the lifespan.shutdown signal is received
        """
        if self._engine:
            await self._engine.dispose()
            self._engine = None

    async def execute_raw(self, statement, *, as_transaction: bool = False) -> list[RowMapping]:
        """
        Executes raw sql statement and returns list of RowMapping objects.
        
        If as_transaction is True, the execution will be wrapped in a transaction.
        as_transaction=False is for read-only state,emts; DML 
        """
        if not self._session_factory:
            raise RuntimeError("Database is not connected.")
        async with self._session_factory() as session:
            if as_transaction:
                async with session.begin():
                    result = await session.execute(statement)
            else:
                result = await session.execute(statement)
            return result.mappings().all()

    @property
    def db_uri(self):
        """
        Returns database connection uri string
        """
        return self._db_uri

    @property
    def engine(self) -> AsyncEngine:
        """
        Returns database engine
        """
        if self._engine is None:
            raise RuntimeError("Engine not initialized. Call connect() first.")
        return cast(AsyncEngine, self._engine)

    @property
    def variable_prefix(self) -> str:
        """
        Return the config variables prefix string
        """
        return self._variable_prefix

    @property
    def db_name(self) -> str:
        return self.__db_name__

    @property
    def managed_session(self) -> Callable:
        """
        Returns a decorator that:
        - Creates a new AsyncSession per request.
        - Injects it into the kwargs of the request with the key "session" or custom session name.
        - Commits if no error occurs.
        - Rolls back if an unhandled error occurs.
        - Closes the session automatically afterward.
        """

        def decorator(handler: Callable) -> Callable:
            @wraps(handler)
            async def wrapper(*args, **kwargs):
                if not self._session_factory:
                    raise RuntimeError(
                        "Database is not connected. "
                        "Connection should be established automatically."
                        "Please check network connection and configurations."
                    )
                async with self._session_factory() as session:  # Ensures session closure
                    async with session.begin():  # Ensures transaction handling (auto commit/rollback)
                        kwargs[self.session_name] = session
                        return await run_sync_or_async(handler, *args, **kwargs)
            return wrapper
        return decorator
    
    @property
    def readonly_session(self) -> Callable:
        """
        Returns a decorator that:
        - Creates a new AsyncSession per request.
        - Injects it into the kwargs of the request with the key "session" or custom session name.
        - Closes the session automatically afterward.
        - Does not commit or rollback, for read-only operations.
        """
        def decorator(handler: Callable) -> Callable:
            @wraps(handler)
            async def wrapper(*args, **kwargs):
                if not self._session_factory:
                    raise RuntimeError(
                        "Database is not connected. "
                        "Connection should be established automatically."
                        "Please check network connection and configurations."
                    )
                async with self._session_factory() as session:  # Ensures session closure
                    kwargs[self.session_name] = session
                    return await run_sync_or_async(handler, *args, **kwargs)
            return wrapper
        return decorator



"""
Exception handler for application
"""
from typing import Any
from pyjolt import Request, Response, HttpStatus
from pyjolt.exceptions import ExceptionHandler, handles
from pydantic import BaseModel, ValidationError

class ErrorResponse(BaseModel):
    message: str
    status: str = "error"
    details: Any|None = None

class Handler(ExceptionHandler):

    @handles(ValidationError)
    async def validation_error_handler(self, req: Request, exc: ValidationError) -> Response[ErrorResponse]:
        """Handles validation errors"""
        #Parses validation error details
        details = {}
        if hasattr(exc, "errors"):
            for error in exc.errors():
                details[error["loc"][0]] = error["msg"]

        return req.response.json({
            "message": "Validation failed.",
            "details": details
        }).status(HttpStatus.UNPROCESSABLE_ENTITY)



"""
Authentication api
"""
from app.api.models import User
from app.authentication import auth
from app.extensions import db
from pydantic import BaseModel

from pyjolt import HttpStatus, MediaType, Request, Response
from pyjolt.controller import Controller, consumes, path, post, produces


class LoginData(BaseModel):

    email: str
    password: str


@path("/api/v1/auth")
class AuthApi(Controller):

    @post("/")
    @consumes(MediaType.APPLICATION_JSON)
    @produces(MediaType.APPLICATION_JSON)
    async def login(self, req: Request, data: LoginData) -> Response:
        session = db.create_session()
        user: User = await User.query(session).filter_by(email=data.email).first()
        if user is None:
            return req.response.json({
                "message": "Wrong credentials",
                "status": "error"
            }).status(HttpStatus.FORBIDDEN)
        cookie: str = auth.create_signed_cookie_value(user.id)
        req.response.set_cookie("auth_cookie", cookie, 86400, http_only=True)
        return req.response.json({
            "message": "Login successful",
            "status": "success"
        }).status(HttpStatus.OK)



"""
Test app implementation
"""

from app.configs import Config
from pyjolt import PyJolt, app, on_shutdown, on_startup


@app(__name__, configs = Config)
class Application(PyJolt):

    @on_startup
    async def first_startup_method(self):
        print("Starting up...", self.url_for("Static.get", filename="image.png"))
        print("Starting up...", self.url_for("static", filename="image.png"))

    @on_shutdown
    async def first_shutdown_method(self):
        print("Shuting down...", self.url_for("Static.get", filename="image.png"))



"""
User models
"""
from sqlalchemy import Integer, String, ForeignKey
from sqlalchemy.orm import mapped_column, Mapped, relationship

from .base_model import BaseModel

class User(BaseModel):
    """
    User model
    """
    __tablename__: str = "users"

    fullname: Mapped[str] = mapped_column(String(30))
    email: Mapped[str] = mapped_column(String(50), unique=True)
    age: Mapped[int] = mapped_column(Integer)
    description: Mapped[str] = mapped_column(String(100), nullable=True)

    roles: Mapped[list["Role"]] = relationship(
        back_populates="user",
        cascade="all, delete-orphan",
        single_parent=True,
        lazy="immediate"
    )

class Role(BaseModel):
    """
    User role
    """
    __tablename__: str = "roles"

    role: Mapped[str] = mapped_column(String(20))
    user_id: Mapped[int] = mapped_column(
        ForeignKey("users.id", ondelete="CASCADE"),
        nullable=True,
        index=True
    )
    user: Mapped["User"] = relationship(back_populates="roles") 



"""
Authentication
"""
from enum import StrEnum
from typing import Optional

from app.api.models import User
from app.extensions import db
from pyjolt import Request
from pyjolt.auth import Authentication

class UserRoles(StrEnum):
    ADMIN = "admin"
    SUPERUSER = "superuser"
    USER = "user"

class Auth(Authentication):

    async def user_loader(self, req: Request) -> Optional[User]:
        """Loads user from the provided cookie"""
        cookie_header = req.headers.get("cookie", "")
        if cookie_header:
            # Split the cookie string on semicolons and equals signs to extract individual cookies
            cookies = dict(cookie.strip().split('=', 1) for cookie in cookie_header.split(';'))
            auth_cookie = cookies.get("auth_cookie")
            if auth_cookie:
                user_id = self.decode_signed_cookie(auth_cookie)
                if user_id:
                    user = await User.query(db.create_session()).filter_by(id=user_id).first()
                    return user
        return None

    async def role_check(self, user: User, roles: list[UserRoles]) -> bool:
        """Checks intersection of user roles and required roles"""
        user_roles = set([role.role for role in user.roles])
        return len(user_roles.intersection(set(roles))) > 0

auth: Auth = Auth()



"""
AI interface for PyJolt app
Makes connecting to LLM's easy
"""
from abc import ABC, abstractmethod
import inspect
from functools import wraps
from typing import (List, Dict, Any, Optional, get_type_hints, Callable)
import docstring_parser
from openai import AsyncOpenAI
from openai.types.chat import ChatCompletion, ChatCompletionMessageToolCall

from ..pyjolt import PyJolt, Request, HttpStatus, Response
from ..utilities import run_sync_or_async
from ..exceptions import BaseHttpException
from ..base_extension import BaseExtension

class ChatContextNotFound(BaseHttpException):

    def __init__(self, msg: str, status_code: int|HttpStatus = HttpStatus.NOT_FOUND):
        super().__init__(msg, status_code=status_code)
        if isinstance(status_code, HttpStatus):
            status_code = status_code.value
        self.status_code = status_code

class FailedToRunAiToolMethod(BaseHttpException):

    def __init__(self, msg: str,
                 method_name: str,
                 *args,
                 status_code: int|HttpStatus = HttpStatus.UNPROCESSABLE_ENTITY,
                 **kwargs):
        super().__init__(msg, status_code=status_code)
        if isinstance(status_code, HttpStatus):
            status_code = status_code.value
        self.status_code = status_code
        self.method_name = method_name
        self.args = args
        self.kwargs = kwargs

class AiInterface(BaseExtension, ABC):
    """
    Main AI interface
    """
    _DEFAULT_CONFIGS = {
        "timeout": 30,
        "temperature": 1.0,
        "model": "gpt-3.5-turbo",
        "api_base_url": "https://api.openai.com/v1",
        "response_format": {
            "type": "json_object"
        },
        "tool_choice": False,
        "max_retries": 0
    }

    def __init__(self, variable_prefix: str = ""):
        """
        Extension init method
        """
        self._app: PyJolt = None
        self._api_key: str = None
        self._api_base_url: str = None
        self._organization_id: str = None
        self._project_id: str = None
        self._variable_prefix: str = variable_prefix
        self._timeout: int = None
        self._model: str = None
        self._temperature: float = None
        self._response_format: dict[str, str] = None
        self._tool_choice = False
        self._max_retries: int = None
        self._tools: list = []
        self._tools_mapping: dict[str, Callable] = {}
        self._chat_session_loader: Callable = None
        self._provider_methods: dict[str, Callable] = {}
        self._chat_session_model_type = None
        self._chat_context_name: str = "chat_context"
        
        self._get_tools()

    def init_app(self, app: PyJolt):
        """
        Initilizer method for extension
        """
        self._app = app
        self._api_key = self._app.get_conf(f"{self._variable_prefix}AI_INTERFACE_API_KEY",
                                           None)
        self._api_base_url = self._app.get_conf(f"{self._variable_prefix}AI_INTERFACE_API_BASE_URL",
                                                self._DEFAULT_CONFIGS["api_base_url"])
        self._organization_id = self._app.get_conf(f"{self._variable_prefix}AI_INTERFACE_ORGANIZATION_ID",
                                                    None)
        self._project_id = self._app.get_conf(f"{self._variable_prefix}AI_INTERFACE_PROJECT_ID",
                                                    None)
        self._timeout = self._app.get_conf(f"{self._variable_prefix}AI_INTERFACE_TIMEOUT",
                                           self._DEFAULT_CONFIGS["timeout"])
        self._model = self._app.get_conf(f"{self._variable_prefix}AI_INTERFACE_MODEL",
                                         self._DEFAULT_CONFIGS["model"])
        self._temperature = self._app.get_conf(f"{self._variable_prefix}AI_INTERFACE_TEMPERATURE",
                                               self._DEFAULT_CONFIGS["temperature"])
        self._response_format = self._app.get_conf(f"{self._variable_prefix}AI_INTERFACE_RESPONSE_FORMAT",
                                                   self._DEFAULT_CONFIGS["response_format"])
        self._tool_choice = self._app.get_conf(f"{self._variable_prefix}AI_INTERFACE_TOOL_CHOICE",
                                               self._DEFAULT_CONFIGS["tool_choice"])
        self._max_retries = self._app.get_conf(f"{self._variable_prefix}AI_INTERFACE_MAX_RETRIES",
                                               self._DEFAULT_CONFIGS["max_retries"])
        self._chat_context_name = self._app.get_conf(f"{self._variable_prefix}AI_INTERFACE_CHAT_CONTEXT_NAME",
                                                    self._chat_context_name)

        self._app.add_extension(self)
    
    @property
    def configs(self) -> dict[str, str|int|float|dict|None]:
        """
        Returns default configs object with env. var. or extension defaults
        """
        return {
            "api_key": self._api_key,
            "api_base_url": self._api_base_url,
            "organization_id": self._organization_id,
            "project_id": self._project_id,
            "timeout": self._timeout,
            "model": self._model,
            "temperature": self._temperature,
            "response_format": self._response_format,
            "tool_choice": self._tool_choice,
            "max_retries": self._max_retries
        }

    async def provider(self,
                    messages: List[Dict[str, str]],
                    **kwargs) -> tuple[str|None, list[ChatCompletionMessageToolCall]|None, ChatCompletion|None]:
        """
        Default provider method. Uses AsyncOpenAI from the openai package
        """

        # Build request
        api_key = kwargs.get("api_key", self._api_key)
        organization = kwargs.get("organization", self._organization_id)
        project = kwargs.get("project", self._project_id)
        timeout = kwargs.get("timeout", self._timeout)
        base_url = kwargs.get("api_base_url", self._api_base_url)
        max_retries = kwargs.get("max_retries", self._max_retries)

        client: AsyncOpenAI = AsyncOpenAI(
            api_key=api_key,
            organization=organization,
            project=project,
            timeout=timeout,
            base_url=base_url,
            max_retries=max_retries
        )

        model = kwargs.get("model", self._model)
        temperature = kwargs.get("temperature", self._temperature)
        response_format = kwargs.get("response_format", self._response_format)

        configs: dict = {
            "messages": messages,
            "model": model,
            "temperature": temperature,
            "response_format": response_format,
        }
        if kwargs.get("use_tools", False):
            configs["tools"] = self._tools

        chat: ChatCompletion = await client.chat.completions.create(**configs)
        tool_calls = chat.choices[0].message.tool_calls or None
        assistant_message_content = chat.choices[0].message.content or None
        return assistant_message_content, tool_calls, chat

    async def create_chat_completion(self,
        messages: List[Dict[str, str]],
        **kwargs) -> tuple[str|None, list[ChatCompletionMessageToolCall]|None, ChatCompletion|None]:
        """
        Makes prompt with chosen provider method.
        Default is the default_provider method which is OpenAI compatible.

        :param stream: bool = False
        :returns chat_completion:
        """
        ##if default method is selected
        return await self.provider(messages, **kwargs)
        
    
    async def envoke_ai_tool(self, tool_name, *args, **kwargs) -> Any:
        """
        Runs a registered AI tool method
        """
        tool_method: Optional[Callable] = self._tools_mapping.get(tool_name, None)
        if tool_method is None:
            raise ValueError(f"Tool method named {tool_name} is not registered with the AI interface")
        try:
            return await run_sync_or_async(tool_method, *args, **kwargs)
        except Exception as exc:
            raise FailedToRunAiToolMethod(f"Failed to run AI tool {tool_name}", *args, **kwargs) from exc

    def build_function_schema(self, func: Callable,
                                    func_name: Optional[str] = None,
                                     description: Optional[str] = None) -> dict[str, any]:
        """
        Automatically builds an OpenAI function schema from a Python function.
        Assumes docstring and type hints follow some basic conventions.
        """
        # Parse the docstring
        doc = docstring_parser.parse(func.__doc__ or "")
        func_description = description or doc.short_description or ""

        # Build the skeleton
        schema: dict[str, Any] = {
            "type": "function",
            "function": {
                "name": func_name or func.__name__,
                "description": func_description,
                "parameters": {
                    "type": "object",
                    "properties": {},
                    "required": [],
                    "additionalProperties": False,
                }
            }
        }

        # Collect parameter info
        sig = inspect.signature(func)
        hints = get_type_hints(func)

        for param_name, param in sig.parameters.items():
            # Derive param type from hints
            param_type = hints.get(param_name, str)
            if param_type is str:
                schema_type = "string"
            elif param_type in [int, float]:
                schema_type = "number"
            elif param_type is bool:
                schema_type = "boolean"
            else:
                schema_type = "string"

            param_desc: str|None = ""
            for doc_param in doc.params:
                if doc_param.arg_name == param_name:
                    param_desc = doc_param.description
                    break

            schema["function"]["parameters"]["properties"][param_name] = {
                "type": schema_type,
                "description": param_desc,
            }
            #If no default value is detected, the parameter is required
            if param.default is inspect.Parameter.empty:
                schema["function"]["parameters"]["required"].append(param_name)
        return schema

    @abstractmethod
    async def chat_context_loader(self, req: Request, *args: Any, **kwargs: Any) -> Optional[Any]:
        """Should load and return a chat session object (ie db model) or none"""

    @property
    def chat_context_name(self) -> str:
        return self._chat_context_name

    @property
    def with_chat_context(self) -> Callable:
        """
        Decorator for injecting chat session to route handler.
        Uses the chat session loader method added with the
        @chat_context_loader decorator.

        Injects the chat context object as a keyword argument
        """
        interface: AiInterface = self
        def decorator(func: Callable):
            @wraps(func)
            async def wrapper(self, *args, **kwargs) -> "Response":
                req: Request = args[0]
                if not isinstance(req, Request):
                    raise ValueError("Missing Request object at @with_chat_context decorator. The request object"
                                     " must be the first argument of the route handler. Please check if you have "
                                     "changed the argument sequence.")
                chat_context = await run_sync_or_async(interface.chat_context_loader, req)
                if chat_context is None:
                    raise ChatContextNotFound("Chat session not found")
                kwargs[interface.chat_context_name] = chat_context
                return await run_sync_or_async(func, self, *args, **kwargs)
            return wrapper
        return decorator

    def _get_tools(self):
        for name in dir(self):
            method = getattr(self, name)
            if not callable(method):
                continue

            is_tool = getattr(method, "__ai_tool", None) or None
            if not is_tool:
                continue
            self._tools.append(self.build_function_schema(method, is_tool["name"],
                                                          is_tool["description"]))
            self._tools_mapping[is_tool["name"]] = method

def tool(name: Optional[str] = None, description: Optional[str] = None):
    """
    Decorator for adding a method as a tool to the Ai interface
    """
    def decorator(func: Callable):
        """
        Marks method to as ai interface tool
        """
        setattr(func, "__ai_tool", {
            "name": name or func.__name__,
            "description": description or func.__doc__
        })
        return func
    return decorator



"""
Task manager class
"""
from typing import Callable, Tuple, Optional, cast, TYPE_CHECKING
from functools import wraps

from apscheduler.job import Job
from apscheduler.schedulers.asyncio import AsyncIOScheduler
from apscheduler.jobstores.base import JobLookupError
from apscheduler.jobstores.memory import MemoryJobStore
from apscheduler.executors.asyncio import AsyncIOExecutor

from ..utilities import run_sync_or_async, run_in_background
from ..base_extension import BaseExtension

if TYPE_CHECKING:
    from ..pyjolt import PyJolt

class TaskManager(BaseExtension):
    """
    Task manager class for scheduling and managing backgroudn tasks.
    """
    _DEFAULT_CONFIGS = {
        "default_scheduler": AsyncIOScheduler,
        "default_jobstores": {
            'default': MemoryJobStore()
        },
        "default_executors": {
            'default': AsyncIOExecutor(),
        },
        "default_job_defaults": {
            'coalesce': False,
            'max_instances': 3
        },
        "default_daemon": True
    }

    

    def __init__(self, variable_prefix: str = "") -> None:
        self._variable_prefix: str = variable_prefix
        self._app: "Optional[PyJolt]" = None
        self._job_stores: Optional[dict] = self._DEFAULT_CONFIGS["default_jobstores"]
        self._executors: Optional[dict] = self._DEFAULT_CONFIGS["default_executors"]
        self._job_defaults: Optional[dict] = self._DEFAULT_CONFIGS["default_job_defaults"]
        self._daemon: bool = self._DEFAULT_CONFIGS["default_daemon"]
        self._scheduler: Optional[AsyncIOScheduler] = self._DEFAULT_CONFIGS["default_scheduler"]
        self._initial_jobs_methods_list: Optional[list[Tuple]] = []
        self._active_jobs: dict[str, Job] = {}
        self._get_defined_jobs()

    def init_app(self, app: "PyJolt"):
        """
        Initlizer for TaskManager with PyJolt app
        """
        self._app = app
        self._job_stores = self._app.get_conf(f"{self._variable_prefix}TASK_MANAGER_JOB_STORES",
                                              self._DEFAULT_CONFIGS["default_jobstores"])
        self._executors = self._app.get_conf(f"{self._variable_prefix}TASK_MANAGER_EXECUTORS",
                                             self._DEFAULT_CONFIGS["default_executors"])
        self._job_defaults = self._app.get_conf(f"{self._variable_prefix}TASK_MANAGER_JOB_DEFAULTS",
                                                self._DEFAULT_CONFIGS["default_job_defaults"])
        self._daemon = self._app.get_conf(f"{self._variable_prefix}TASK_MANAGER_DAEMON",
                                          self._DEFAULT_CONFIGS["default_daemon"])
        self._scheduler = self._app.get_conf(f"{self._variable_prefix}TASK_MANAGER_SCHEDULER",
                                             self._DEFAULT_CONFIGS["default_scheduler"])
        self._scheduler = self._scheduler(jobstores=self._job_stores,
                                            executors=self._executors,
                                            job_defaults=self._job_defaults,
                                            daemon=self._daemon
                                            )
        self._app.add_extension(self)
        self._app.add_on_startup_method(self._start_scheduler)
        self._app.add_on_shutdown_method(self._stop_scheduler)

    def start_scheduler(self):
        """
        Starts the scheduler
        """
        self._start_scheduler(None)

    def stop_scheduler(self):
        """
        Stop the scheduler
        """
        self._stop_scheduler(None)
        self._active_jobs = None

    def pause_scheduler(self):
        """
        Pauses scheduler execution
        """
        self.scheduler.pause()

    def resume_scheduler(self):
        """
        Resumes paused scheduler execution
        """
        self.scheduler.resume()

    async def _start_scheduler(self):
        """
        On startup hook for starting the scheduler
        """
        self.scheduler.start()
        self._start_initial_jobs()

    async def _stop_scheduler(self):
        """
        On shutdown hook for shuting the scheduler down
        """
        self.scheduler.shutdown()
    
    def _get_defined_jobs(self):
        for name in dir(self):
            method = getattr(self, name)
            if not callable(method):
                continue
            scheduler_method = getattr(method, "_scheduler_job", None)
            if scheduler_method:
                self._initial_jobs_methods_list.append((method,
                                                        scheduler_method["args"],
                                                        scheduler_method["kwargs"]))

    def _start_initial_jobs(self) -> None:
        """
        Starts all initial jobs (decorated functions)
        """
        if self._initial_jobs_methods_list is None:
            return
        for func, args, kwargs in self._initial_jobs_methods_list:
            job: Job = self.scheduler.add_job(func, *args, **kwargs)
            self._active_jobs[job.id] = job
        self._initial_jobs_methods_list = None

    def run_background_task(self, func: Callable, *args, **kwargs):
        """
        Runs a method in the background (fire and forget).
        Used for running function whose execution doesn't have to be awaited.
        Example: a route handler can return a response immediately, and the
        task is executed in a seperate thread (sending an email for example).

        Uses the pyjolt.utilities.run_in_background method
        """
        run_in_background(func, *args, **kwargs)

    def add_job(self, func: Callable, *args, **kwargs) -> Job:
        """
        Adds job
        """
        job: Job = self.scheduler.add_job(func, *args, **kwargs)
        self._active_jobs[job.id] = job
        return job

    def remove_job(self, job: str|Job, job_store: Optional[str] = None):
        """
        Removes a job.
        :param job: job id (str) or the Job instance returned by the scheduler.add_job method
        """
        if isinstance(job, Job):
            job = job.id
        return self._remove_job(job, job_store)

    def pause_job(self, job: str|Job):
        """
        Pauses the job
        """
        if isinstance(job, Job):
            return job.pause()
        active_job: Job = self._active_jobs.get(job, None)
        if job is None:
            raise JobLookupError(job)
        return active_job.pause()

    def resume_job(self, job: str|Job):
        """
        Resumes job
        :param paused_job: id or Job instance
        """
        if isinstance(job, Job):
            return job.resume()
        paused_job: Job = self._active_jobs.get(job, None)
        if paused_job is None:
            raise JobLookupError(paused_job)
        return paused_job.resume()
    
    def get_job(self, job_id: str) -> Job|None:
        return self._active_jobs.get(job_id, None)

    def _remove_job(self, job_id: str, job_store = None):
        """
        Removes job from job list
        """
        self.scheduler.remove_job(job_id, job_store)
        del self._active_jobs[job_id]

    @property
    def jobs(self) -> dict[str, Job]:
        """
        Returns list of running jobs
        """
        return self._active_jobs

    @property
    def scheduler(self) -> AsyncIOScheduler:
        """
        Returns the background scheduler instance
        """
        return self._scheduler
    
    @property
    def app(self) -> "PyJolt":
        """
        Application instance
        """
        return cast("PyJolt", self._app)


def schedule_job(*args, **kwargs):
    """
    ```
    A decorator to add a function as a scheduled job in the given APScheduler instance.
    The decorated function is added to a list of tuples (func, args, kwargs) and the job
    is started when the scheduler instance is started (on_startup event of PyJolt)
    IMPORTANT: The decorator should be the top-most decorator of the function to make sure
    any other decorator is applied before the job is added to the job list
    :param args: Positional arguments to pass to scheduler.add_job().
                Typically, the first of these args is the trigger (e.g. 'interval', 'cron', etc.).
    :param kwargs: Keyword arguments to pass to scheduler.add_job().
    Example:
    Runs a job with id 'my_job_id' every 5 minutes

    @schedule_job('interval', minutes=5, id='my_job_id')
    async def my_job(self):
        #some task
    ```
    """
    def decorator(func: Callable):
        @wraps(func)
        async def wrapper(self, *f_args, **f_kwargs):
            return await run_sync_or_async(func, self, *f_args, **f_kwargs)
        setattr(wrapper, "_scheduler_job", {"args": args, "kwargs": kwargs})
        return wrapper
    return decorator



"""Custom api exceptions"""
from pyjolt.exceptions import BaseHttpException
from pyjolt import HttpStatus

class EntityNotFound(BaseHttpException):
    """Exception for entity not found"""

    def __init__(self, message: str, status_code: HttpStatus = HttpStatus.NOT_FOUND):
        super().__init__(message=message, status_code=status_code)



"""
AI Interface
"""
from typing import Optional

from app.api.models.chat_session import ChatSession
from app.extensions import db

from pyjolt.database import AsyncSession
from pyjolt.ai_interface import AiInterface
from pyjolt.request import Request


class Interface(AiInterface):

    @db.managed_session
    #pylint: disable-next=W0221
    async def chat_context_loader(self, req: Request,
                                  session: AsyncSession) -> Optional[ChatSession]:
        chat_session_id: Optional[int] = req.route_parameters.get("chat_session_id",
                                                                  None)
        if chat_session_id is None:
            return None
        return await ChatSession.query(session).filter_by(id = chat_session_id).first()

ai_interface: Interface = Interface()



"""
Exception handler api
"""
from typing import Any
from pydantic import BaseModel, ValidationError
from pyjolt.exceptions import ExceptionHandler, handles, AuthenticationException, UnauthorizedException, NotFound
from pyjolt import Request, Response, HttpStatus

from .custom_exceptions import EntityNotFound

class ErrorResponse(BaseModel):
    message: str
    details: Any|None = None

class CustomExceptionHandler(ExceptionHandler):

    @handles(EntityNotFound)
    async def not_found(self, req: "Request", exc: EntityNotFound) -> "Response[ErrorResponse]":
        """Handles not found exceptions"""
        return req.response.json({
            "message": exc.message
        }).status(exc.status_code)
    
    @handles(ValidationError)
    async def validation_error(self, req: "Request", exc: ValidationError) -> "Response[ErrorResponse]":
        """Handles validation errors"""
        details = {}
        if hasattr(exc, "errors"):
            for error in exc.errors():
                details[error["loc"][0]] = error["msg"]

        return req.response.json({
            "message": "Validation failed.",
            "details": details
        }).status(HttpStatus.UNPROCESSABLE_ENTITY)
    
    @handles(AuthenticationException)
    async def auth_exception(self, req: "Request", exc: AuthenticationException) -> "Response[ErrorResponse]":
        """Handles authentication errors"""
        return req.response.json({
            "message": exc.message,
            "details": "error"
        }).status(exc.status_code)
    
    @handles(UnauthorizedException)
    async def unauthorized_exception(self, req: "Request", exc: UnauthorizedException) -> "Response[ErrorResponse]":
        """Handled unauthorized access errors"""
        return req.response.json({
            "message": exc.message,
            "details": exc.data,
        }).status(exc.status_code)
    
    @handles(NotFound)
    async def route_not_found(self, req: "Request", exc: NotFound) -> "Response[ErrorResponse]":
        """Handles NotFound exception - from Werkzeug"""
        return req.response.json({
            "message": "Endpoint was not found",
            "details": "Not found"
        }).status(HttpStatus.NOT_FOUND)



"""
A CLI controller example. You can register mutiple controllers and add as many methods as you like to individual controllers.
To add a method to a controller, use the @command decorator to define the command name and help text, and the (multiple) @argument decorator(s) to define any arguments the command takes.
"""

from pyjolt.cli import CLIController, command, argument

class UtilityCLIController(CLIController):
    """A simple CLI utility controller."""

    @command("greet", help="Greet a user with a message.")
    @argument("name", arg_type=str, description="The name of the user to greet.")
    async def greet(self, name: str):
        """Greet by name."""
        print(f"Hello, {name}! Welcome to the CLI utility.")

    @command("add", help="Add two numbers.")
    @argument("a", arg_type=int, description="The first number.")
    @argument("b", arg_type=int, description="The second number.")
    async def add(self, a: int, b: int):
        """Add two numbers and print the result."""
        result = a + b
        print(f"The sum of {a} and {b} is {result}.")



"""
authentication.py
Authentication module of PyJolt
"""
from abc import ABC, abstractmethod
from typing import Callable, Optional, Dict, Any, TYPE_CHECKING, cast
from functools import wraps
import base64
from datetime import datetime, timedelta, timezone

import bcrypt
import jwt
import binascii
from cryptography.hazmat.primitives.hmac import HMAC
from cryptography.hazmat.primitives import hashes
from cryptography.exceptions import InvalidSignature

from ..exceptions import AuthenticationException, UnauthorizedException
from ..utilities import run_sync_or_async
from ..request import Request
from ..base_extension import BaseExtension
if TYPE_CHECKING:
    from ..pyjolt import PyJolt
    from ..response import Response
    from ..controller import Controller

class Authentication(BaseExtension, ABC):
    """
    Authentication class for PyJolt
    """

    REQUEST_ARGS_ERROR_MSG: str = ("Injected argument 'req' of route handler is not an instance "
                    "of the Request class. If you used additional decorators "
                    "make sure the order of arguments was not changed. "
                    "The Request argument must always come first.")
    
    USER_LOADER_ERROR_MSG: str = ("Undefined user loader method. Please define a user loader "
                                  "method with the @user_loader decorator before using "
                                  "the login_required decorator")

    _DEFAULT_CONFIGS = {
        "DEFAULT_AUTHENTICATION_ERROR_MESSAGE": "Login required",
        "DEFAULT_AUTHORIZATION_ERROR_MESSAGE": "Missing user role(s)"
    }

    def __init__(self, variable_prefix: str = "") -> None:
        """
        Initilizer for authentication module
        """
        self._app: "Optional[PyJolt]" = None
        self._variable_prefix: str = variable_prefix
        self.authentication_error: str = self._DEFAULT_CONFIGS["DEFAULT_AUTHENTICATION_ERROR_MESSAGE"]
        self.authorization_error: str = self._DEFAULT_CONFIGS["DEFAULT_AUTHORIZATION_ERROR_MESSAGE"]

    def init_app(self, app: "PyJolt"):
        """
        Configures authentication module
        """
        self._app = app
        self.authentication_error = app.get_conf(f"{self._variable_prefix}AUTHENTICATION_ERROR_MESSAGE",
                                                 self.authentication_error)
        self.authorization_error = app.get_conf(f"{self._variable_prefix}UNAUTHORIZED_ERROR_MESSAGE",
                                                 self.authorization_error)
        self._app.add_extension(self)

    def create_signed_cookie_value(self, value: str|int) -> str:
        """
        Creates a signed cookie value using HMAC and a secret key.

        value: The string value to be signed
        secret_key: The application's secret key for signing

        Returns a base64-encoded signed value.
        """
        if isinstance(value, int):
            value = f"{value}"

        hmac_instance = HMAC(self.secret_key.encode("utf-8"), hashes.SHA256())
        hmac_instance.update(value.encode("utf-8"))
        signature = hmac_instance.finalize()
        signed_value = f"{value}|{base64.urlsafe_b64encode(signature).decode('utf-8')}"
        return signed_value

    def decode_signed_cookie(self, cookie_value: str) -> str:
        """
        Decodes and verifies a signed cookie value.

        cookie_value: The signed cookie value to be verified and decoded
        secret_key: The application's secret key for verification

        Returns the original string value if the signature is valid.
        Raises a ValueError if the signature is invalid.
        """
        try:
            value, signature = cookie_value.rsplit("|", 1)
            signature_bytes = base64.urlsafe_b64decode(signature)
            hmac_instance = HMAC(self.secret_key.encode("utf-8"), hashes.SHA256())
            hmac_instance.update(value.encode("utf-8"))
            hmac_instance.verify(signature_bytes)  # Throws an exception if invalid
            return value
        except (ValueError, IndexError, binascii.Error, InvalidSignature):
            # pylint: disable-next=W0707
            raise ValueError("Invalid signed cookie format or signature.")

    def create_password_hash(self, password: str) -> str:
        """
        Creates a secure hash for a given password.

        password: The plain text password to be hashed
        Returns the hashed password as a string.
        """
        hashed = bcrypt.hashpw(password.encode("utf-8"), bcrypt.gensalt())
        return hashed.decode("utf-8")

    def check_password_hash(self, password: str, hashed_password: str) -> bool:
        """
        Verifies a given password against a hashed password.

        password: The plain text password provided by the user
        hashed_password: The stored hashed password
        Returns True if the password matches, False otherwise.
        """
        return bcrypt.checkpw(password.encode("utf-8"), hashed_password.encode("utf-8"))
    
    def create_jwt_token(self, payload: Dict, expires_in: int = 3600) -> str:
        """
        Creates a JWT token.

        :param payload: A dictionary containing the payload data.
        :param expires_in: Token expiry time in seconds (default: 3600 seconds = 1 hour).
        :return: Encoded JWT token as a string.
        """
        if not isinstance(payload, dict):
            raise ValueError("Payload must be a dictionary.")

        # Add expiry to the payload
        payload = payload.copy()
        payload["exp"] = datetime.now(timezone.utc) + timedelta(seconds=expires_in)

        # Create the token using the app's SECRET_KEY
        token = jwt.encode(payload, self.secret_key, algorithm="HS256")
        return token

    def validate_jwt_token(self, token: str) -> Dict|None:
        """
        Validates a JWT token.

        :param token: The JWT token to validate.
        :return: Decoded payload if the token is valid.
        :raises: InvalidJWTError if the token is expired.
                 InvalidJWTError for other validation issues.
        """
        try:
            # Decode the token using the app's SECRET_KEY
            payload = jwt.decode(token, self.secret_key, algorithms=["HS256"])
            return payload
        except (jwt.ExpiredSignatureError, jwt.InvalidTokenError):
            return None
    
    @property
    def secret_key(self):
        """
        Returns app secret key or none
        """
        sec_key = self._app.get_conf("SECRET_KEY", None)
        if sec_key is None:
            raise ValueError("SECRET_KEY is not defined in app configurations")
        return sec_key

    @property
    def login_required(self) -> Callable[[Callable], Callable]:
        """
        Decorator enforcing that a user is authenticated before the endpoint runs.

        Usage:
            @auth.login_required
            async def endpoint(self, req: Request, ...): ...
        """
        authenticator = self

        def decorator(handler: Callable) -> Callable:
            @wraps(handler)
            async def wrapper(self: "Controller", *args, **kwargs) -> "Response":
                if not args:
                    raise RuntimeError(
                        "Request must be auto-injected as the first argument after self."
                    )
                req: "Request" = args[0]
                if not isinstance(req, Request):
                    raise ValueError(authenticator.REQUEST_ARGS_ERROR_MSG)
                if req.user is None:
                    user_loader = getattr(authenticator, "user_loader", None)
                    if user_loader is None:
                        raise ValueError(authenticator.USER_LOADER_ERROR_MSG)
                    req.set_user(await run_sync_or_async(user_loader, req))
                if req.user is None:
                    # Not authenticated
                    raise AuthenticationException(authenticator.authentication_error)

                return await run_sync_or_async(handler, self, *args, **kwargs)

            return wrapper

        return decorator
    

    def role_required(self, *roles) -> Callable[[Callable], Callable]:
        """
        Decorator enforcing that a user has designated roles.
        Decorator must be BELOW the login_required decorator
        Usage:
            @auth.role_required(*roles)
            async def endpoint(self, req: Request, ...): ...
        """
        authenticator = self

        def decorator(handler: Callable) -> Callable:
            @wraps(handler)
            async def wrapper(self: "Controller", *args, **kwargs) -> "Response":
                if not args:
                    raise RuntimeError(
                        "Request must be auto-injected as the first argument after self."
                    )
                req: "Request" = args[0]
                if not isinstance(req, Request):
                    raise ValueError(authenticator.REQUEST_ARGS_ERROR_MSG)

                if req.user is None:
                    raise RuntimeError(
                        "User not loaded. Make sure the method is decorated with @login_required to load the user object"
                    )
                authorized: bool = await run_sync_or_async(authenticator.role_check, req.user, list(roles))
                if not authorized:
                    #not authorized
                    raise UnauthorizedException(authenticator.authorization_error, list(roles))

                return await run_sync_or_async(handler, self, *args, **kwargs)

            return wrapper

        return decorator

    @abstractmethod
    async def user_loader(self, req: "Request") -> Any:
        """
        Should return a user object (or None) loaded from the cookie
        or some other way provided by the request object
        """

    @abstractmethod
    async def role_check(self, user: Any, roles: list[Any]) -> bool:
        """
        Should check if user has required role(s) and return a boolean
        True -> user has role(s)
        False -> user doesn't have role(s)
        """



"""
Base model
"""
from sqlalchemy.orm import mapped_column, Mapped

from pyjolt.database import create_declerative_base

Base = create_declerative_base("db")

class BaseModel(Base):

    __abstract__ = True
    id: Mapped[int] = mapped_column(primary_key=True)



"""
PyJolt cli
"""
from typing import Callable
from pathlib import Path

from .new_project import new_project

methods: dict[str, Callable] = {
    "new-project": new_project
}
import argparse

def main():
    parser = argparse.ArgumentParser(prog="pyjolt")
    subparsers = parser.add_subparsers(dest="command")

    new_project_parser = subparsers.add_parser("new-project")
    new_project_parser.add_argument("--name")

    args = parser.parse_args()
    method = methods.get(args.command, None)
    if method is not None:
        args = vars(args)
        del args["command"]
        return method(Path.cwd(), **args)



"""
New project initilizer
"""
import shutil
import os
from pathlib import Path

def new_project(cwd: str, name: str = None) -> None:
    if name is None:
        name = input("Project name: ")
    templates_path: Path = Path(os.path.join(os.path.dirname(__file__), "templates", "new_project"))
    destination_path: Path = Path(os.path.join(cwd, name))
    print(f"Creating new project in {destination_path}")

    destination_path.mkdir(parents=True, exist_ok=True)

    for item in templates_path.iterdir():
        target = destination_path / item.name
        if item.is_dir():
            shutil.copytree(item, target, dirs_exist_ok=True)
        else:
            shutil.copy2(item, target)

